(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Check_DPP
imports Check_Common 
begin

subsection {*Proving finiteness in the QDP Framework*}

definition missing :: "string \<Rightarrow> shows \<Rightarrow> shows" where
  "missing s x \<equiv> ''the '' +#+ s +#+ '' '' +#+ x +@+ shows_string '' is missing''"

definition toomuch :: "string \<Rightarrow> shows \<Rightarrow> shows" where
  "toomuch s x \<equiv> ''superfluous '' +#+ s +#+ '' '' +#+ x +@+ x"

fun
  check_dpp_subsumes ::
    "('d, ('f::show, 'l::show) lab, 'v::show) dpp_ops
    \<Rightarrow> ('f, 'l, 'v) dppLL \<Rightarrow> 'd \<Rightarrow> shows check"
where
  "check_dpp_subsumes I (nfs', m', p', pw', q', r', rw') d = (do {
    let p   = dpp_ops.P I d;
    let pw  = dpp_ops.Pw I d;
    let q   = dpp_ops.Q I d;
    let r   = dpp_ops.R I d;
    let rw  = dpp_ops.Rw I d;
    let nfs = dpp_ops.nfs I d;
    let m   = dpp_ops.minimal I d;
    let pb' = p' @ pw';
    let rb  = r  @ rw;
    let rb' = r' @ rw';
    check (m = m') (shows ''incompatible minimality flags'');
    check (nfs = nfs') (shows ''incompatible substitutions-in-normal-form flags'');
    check_subseteq p p'    <+? (\<lambda>r. toomuch ''pair'' (shows_rule r));
    check_subseteq pw pb'  <+? (\<lambda>r. toomuch ''weak pair'' (shows_rule r));
    check_NF_terms_eq q q' <+? (\<lambda>r. shows ''NF(Q) differs due to term '' +@+ shows_term r);
    check_subseteq r r'    <+? (\<lambda>r. toomuch ''strict rule'' (shows_rule r));
    check_subseteq rb rb'  <+? (\<lambda>r. toomuch ''strict/weak rule'' (shows_rule r));
    check_subseteq rb' rb  <+? (\<lambda>r. missing ''strict/weak rule'' (shows_rule r));
    succeed
  }) <+? (\<lambda>s. ''finiteness of the problem'' +#+ shows_nl +@+ shows_dpp I d +@+ shows_nl
           +@+ ''may not be concluded from assuming finiteness of the problem'' +#+ shows_nl
           +@+ shows_dpp I (dpp_ops.mk I nfs' m' p' pw' q' r' rw') +@+ shows_nl
           +@+ s +@+ shows_nl)"

lemma check_dpp_subsumesE [elim]:
  assumes 1: "dpp_spec I"
    and 2: "isOK (check_dpp_subsumes I (nfs, m, p, pw, q, r, rw) d)"
    and 3: "dpp_ops.minimal I d = m
    \<Longrightarrow> dpp_ops.nfs I d = nfs
    \<Longrightarrow> set (dpp_ops.P I d) \<subseteq> set p
    \<Longrightarrow> set (dpp_ops.Pw I d) \<subseteq> set p \<union> set pw
    \<Longrightarrow> NF_terms (set (dpp_ops.Q I d)) = NF_terms (set q)
    \<Longrightarrow> set (dpp_ops.R I d) \<subseteq> set r
    \<Longrightarrow> set (dpp_ops.R I d) \<union> set (dpp_ops.Rw I d) = set r \<union> set rw
    \<Longrightarrow> P"
  shows "P"
  by (rule 3, insert 1 2, auto simp: Let_def)

datatype ('f, 'l, 'v) dp_termination_proof =
  P_is_Empty
| Subterm_Criterion_Proc "('f, 'l) lab projL" "('f, 'l, 'v) rseqL"
    "('f, 'l, 'v) trsLL" "('f, 'l, 'v) dp_termination_proof"
| Redpair_Proc "('f,'l)lab root_redtriple_impl + ('f, 'l) lab redtriple_impl" "('f, 'l, 'v) trsLL"  "('f, 'l, 'v) dp_termination_proof"
| Redpair_UR_Proc "('f, 'l) lab root_redtriple_impl + ('f, 'l) lab redtriple_impl"
    "('f, 'l, 'v) trsLL" "('f, 'l, 'v) trsLL" "('f, 'l, 'v) dp_termination_proof" 
| Usable_Rules_Proc "('f, 'l, 'v) trsLL" "('f, 'l, 'v) dp_termination_proof" 
| Dep_Graph_Proc "(('f, 'l, 'v) dp_termination_proof option \<times> ('f, 'l, 'v) trsLL) list"
| Mono_Redpair_Proc "('f, 'l) lab redtriple_impl"
    "('f, 'l, 'v) trsLL" "('f, 'l, 'v) trsLL" "('f, 'l, 'v) dp_termination_proof" 
| Mono_Redpair_UR_Proc "('f, 'l) lab redtriple_impl"
    "('f, 'l, 'v) trsLL" "('f, 'l, 'v) trsLL" "('f, 'l, 'v) trsLL" "('f, 'l, 'v) dp_termination_proof"
| Size_Change_Subterm_Proc "((('f, 'l) lab, 'v) rule \<times> ((nat \<times> nat) list \<times> (nat \<times> nat) list)) list"
| Size_Change_Redpair_Proc "('f, 'l) lab redtriple_impl" "('f, 'l, 'v) trsLL option"
    "((('f, 'l) lab, 'v) rule \<times> ((nat \<times> nat) list \<times> (nat \<times> nat) list)) list"
| Uncurry_Proc "nat option" "(('f, 'l) lab, 'v) uncurry_info"
    "('f, 'l, 'v) trsLL" "('f, 'l, 'v) trsLL" "('f, 'l, 'v) dp_termination_proof"
| Fcc_Proc "('f, 'l) lab" "(('f, 'l) lab, 'v) ctxt list"
    "('f, 'l, 'v) trsLL" "('f, 'l, 'v) trsLL" "('f, 'l, 'v) dp_termination_proof"
| Split_Proc
    "('f, 'l, 'v) trsLL" "('f, 'l, 'v) trsLL" 
    "('f, 'l, 'v) dp_termination_proof" "('f, 'l, 'v) dp_termination_proof"
| Semlab_Proc
    "(('f, 'l) lab, 'v) sl_variant" "('f, 'l, 'v) trsLL"
    "(('f, 'l) lab, 'v) term list" "('f, 'l, 'v) trsLL"
    "('f, 'l, 'v) dp_termination_proof"
| Switch_Innermost_Proc "('f, 'l) lab join_info" "('f, 'l, 'v) dp_termination_proof"
| Rewriting_Proc
    "('f, 'l, 'v) trsLL option" "('f, 'l, 'v) ruleLL" "('f, 'l, 'v) ruleLL"
    "('f, 'l, 'v) ruleLL" "('f, 'l, 'v) ruleLL" pos "('f, 'l, 'v) dp_termination_proof"
| Instantiation_Proc "('f, 'l, 'v) ruleLL" "('f, 'l, 'v) trsLL" "('f, 'l, 'v) dp_termination_proof"
| Forward_Instantiation_Proc
    "('f, 'l, 'v) ruleLL" "('f, 'l, 'v) trsLL" "('f, 'l, 'v) trsLL option" "('f, 'l, 'v) dp_termination_proof"
| Narrowing_Proc "('f, 'l, 'v) ruleLL" pos "('f, 'l, 'v) trsLL" "('f, 'l, 'v) dp_termination_proof"
| Assume_Finite
    "('f, 'l, 'v) dppLL" "(('f, 'l, 'v) dppLL \<times> ('f, 'l, 'v) dp_termination_proof) list"
| Unlab_Proc "('f, 'l, 'v) trsLL" "('f, 'l, 'v) trsLL" "('f, 'l, 'v) dp_termination_proof"
| Q_Reduction_Proc "('f, 'l, 'v) termsLL" "('f, 'l, 'v) dp_termination_proof"
| Complex_Constant_Removal_Proc "(('f,'l)lab,'v)complex_constant_removal_prf" "('f, 'l, 'v) dp_termination_proof"
| General_Redpair_Proc
    "('f, 'l) lab redtriple_impl" "('f, 'l, 'v) trsLL" "('f, 'l, 'v) trsLL"
    "(('f, 'l) lab, 'v) cond_red_pair_prf" "('f, 'l, 'v) dp_termination_proof list" 

fun
  get_fcc_option ::
    "('f, 'l, 'v) dp_termination_proof \<Rightarrow>
    (('f, 'l) lab \<times> (('f, 'l) lab, 'v) ctxt list \<times> ('f, 'l, 'v) trsLL \<times> ('f, 'l, 'v) trsLL \<times> ('f, 'l, 'v) dp_termination_proof)option"
where
  "get_fcc_option (Fcc_Proc f fcs Pb' Rb' prf) = Some (f, fcs, Pb', Rb', prf)" |
  "get_fcc_option _ = None"

fun mk_dpp where
  "mk_dpp I (nfs, m, p, pw, q, r, rw) = dpp_ops.mk I nfs m p pw q r rw"

text {*The first argument is a flag that specifies whether partial proofs
(i.e, containing assumptions) are accepted or not.*}
function
  check_dp_termination_proof ::
    "bool \<Rightarrow>
    shows \<Rightarrow>
    ('dpp, ('f::{show,key,countable}, label_type) lab, string) dpp_ops \<Rightarrow>
    'dpp \<Rightarrow>
    ('f, label_type, string) dp_termination_proof \<Rightarrow>
    shows check"
where
  "check_dp_termination_proof assms i I dpp P_is_Empty = debug (i []) ''P is empty''
     (if dpp_ops.P I dpp = [] \<and> (dpp_ops.Pw I dpp = [] \<or> dpp_ops.R I dpp = []) 
          then succeed
          else error (i +@+ '': P is not empty in the following DP-problem '' +#+
         shows_nl +@+ shows_dpp I dpp))"
| "check_dp_termination_proof assms i I dpp (Subterm_Criterion_Proc p rseq rP prf) =
    debug (i []) ''Subterm_Criterion_Proc'' (do {
      let P  = dpp_ops.pairs I dpp;
      let Pr = ceta_list_diff P rP;
      dpp' \<leftarrow> subterm_criterion_proc I p rseq Pr dpp
        <+? (\<lambda>s. i +@+ '': error when applying the subterm criterion to the DP problem''
          +#+ shows_nl +@+ shows_dpp I dpp +@+ shows_nl +@+
          ''and trying to remove the pairs '' +#+ shows_nl +@+ shows_rules Pr
          +@+ shows_nl +@+ s);
      check_dp_termination_proof assms (i +@+ shows_string ''.1'') I dpp' prf
        <+? (\<lambda>s. i +@+ '': error below the subterm criterion'' +#+ shows_nl
          +@+ indent s)
    })"
| "check_dp_termination_proof assms i I dpp (Redpair_Proc redp rP prf) = debug (i []) ''Redpair_Proc'' (do {
         let P = dpp_ops.pairs I dpp;
         let Pr = ceta_list_diff P rP;
         let Proc = (case redp of Inr rp \<Rightarrow> generic_ur_af_redtriple_proc I (get_redtriple rp) None
                        | Inl rrp \<Rightarrow> generic_ur_af_root_redtriple_proc I (get_root_redtriple rrp) None);
         dpp' \<leftarrow> Proc Pr dpp 
           <+? (\<lambda>s. i +@+ '': error when applying the reduction pair processor to remove from the DP problem''
             +#+ shows_nl +@+ shows_dpp I dpp +@+ shows_nl +@+
             '' the pairs '' +#+ shows_nl +@+ shows_rules Pr +@+ shows_nl +@+ s);
         check_dp_termination_proof assms (i +@+ shows_string ''.1'') I dpp' prf
           <+? (\<lambda>s. i +@+ '': error below the reduction pair processor'' +#+ shows_nl
             +@+ indent s)
   })" 
| "check_dp_termination_proof assms i I dpp (Usable_Rules_Proc U prf) = debug (i []) ''Usable_Rules_Proc'' (do {
         dpp' \<leftarrow> usable_rules_proc I U dpp 
           <+? (\<lambda>s. i +@+ '': error when applying the usable rules processor to restrict the DP problem''
             +#+ shows_nl +@+ shows_dpp I dpp +@+ shows_nl +@+
             '' to the usable rules '' +#+ shows_nl +@+ shows_rules U +@+ shows_nl +@+ s);
         check_dp_termination_proof assms (i +@+ shows_string ''.1'') I dpp' prf
           <+? (\<lambda>s. i +@+ '': error below the usable rules processor'' +#+ shows_nl
             +@+ indent s)
   })"
| "check_dp_termination_proof assms i I dpp (Q_Reduction_Proc Q prf) = debug (i []) ''Q_Reduction_Proc'' (do {
         dpp' \<leftarrow> q_reduction_proc I Q dpp 
           <+? (\<lambda>s. i +@+ shows '': error when applying the Q-reduction processor'' +@+ shows_nl +@+ s);
         check_dp_termination_proof assms (i +@+ shows_string ''.1'') I dpp' prf
           <+? (\<lambda>s. i +@+ '': error below the Q-reduction processor'' +#+ shows_nl
             +@+ indent s)
   })"
|  "check_dp_termination_proof assms i I dpp (Mono_Redpair_Proc redp rP rR prf) = debug (i []) ''Mono_Redpair_Proc'' (do {
         let P = dpp_ops.pairs I dpp;
         let Pr = ceta_list_diff P rP;
         let R = dpp_ops.rules I dpp;
         let Rr = ceta_list_diff R rR;
         dpp' \<leftarrow> mono_redpair_proc I (get_redtriple redp) Pr Rr dpp 
           <+? (\<lambda>s. i +@+ '': error when applying the mono reduction pair processor to remove from the DP problem''
             +#+ shows_nl +@+ shows_dpp I dpp +@+ shows_nl +@+
             '' the pairs '' +#+ shows_nl +@+ shows_rules Pr +@+ shows_nl +@+ 
             '' and the rules '' +#+ shows_nl +@+ shows_rules Rr +@+ shows_nl +@+ s);
         check_dp_termination_proof assms (i +@+ shows_string ''.1'') I dpp' prf
           <+? (\<lambda>s. i +@+ '': error below the mono reduction pair processor'' +#+ shows_nl
             +@+ indent s)
   })" 
| "check_dp_termination_proof assms i I dpp (Dep_Graph_Proc edpts) =
     debug (i []) ''Dep_Graph_Proc'' (do {
        pdpps \<leftarrow> dep_graph_proc I dpp edpts
         <+? (\<lambda>s. i +@+ '': error while trying to perform SCC-decomposition  on '' +#+
           shows_nl +@+ shows_dpp I dpp +@+ shows_nl +@+ s);
        check_allm_index (\<lambda> (prof,dpp') j. check_dp_termination_proof assms (i +@+ ''.'' +#+ shows (Suc j)) I dpp' prof) pdpps
          <+? (\<lambda>s. i +@+ '': error below the dependency graph processor'' +#+ shows_nl
            +@+ indent s)
  })"
| "check_dp_termination_proof assms i I dpp (Redpair_UR_Proc redp rP ur prf) = debug (i []) ''Redpair_UR_Proc'' (do {
         let P = dpp_ops.pairs I dpp;
         let Pr = ceta_list_diff P rP;
         let Proc = (case redp of Inr rp \<Rightarrow> generic_ur_af_redtriple_proc I (get_redtriple rp) (Some ur) 
                        | Inl rrp \<Rightarrow> generic_ur_af_root_redtriple_proc I (get_root_redtriple rrp) (Some ur));
         dpp' \<leftarrow> Proc Pr dpp 
           <+? (\<lambda>s. i +@+ '': error when applying the reduction pair processor with usable rules to remove from the DP problem''
             +#+ shows_nl +@+ shows_dpp I dpp +@+ shows_nl +@+
             '' the pairs '' +#+ shows_nl +@+ shows_rules Pr +@+ shows_nl +@+ s);
         check_dp_termination_proof assms (i +@+ shows_string ''.1'') I dpp' prf
           <+? (\<lambda>s. i +@+ '': error below the reduction pair processor'' +#+ shows_nl
             +@+ indent s)
   })" 
|  "check_dp_termination_proof assms i I dpp (Mono_Redpair_UR_Proc redp rP rR ur prf) = debug (i []) ''Mono_Redpair_UR_Proc'' (do {
         let P = dpp_ops.pairs I dpp;
         let Pr = ceta_list_diff P rP;
         let R = dpp_ops.rules I dpp;
         let Rr = ceta_list_diff R rR;
         dpp' \<leftarrow> mono_ur_redpair_proc I (get_redtriple redp) Pr Rr ur dpp 
           <+? (\<lambda>s. i +@+ '': error when applying the mono reduction pair processor with usable rules to remove from the DP problem''
             +#+ shows_nl +@+ shows_dpp I dpp +@+ shows_nl +@+
             '' the pairs '' +#+ shows_nl +@+ shows_rules Pr +@+ shows_nl +@+ 
             '' and the rules '' +#+ shows_nl +@+ shows_rules Rr +@+ shows_nl +@+ s);
         check_dp_termination_proof assms (i +@+ shows_string ''.1'') I dpp' prf
           <+? (\<lambda>s. i +@+ '': error below the mono reduction pair processor with usable rules'' +#+ shows_nl
             +@+ indent s)
   })" 
| "check_dp_termination_proof assms i I dpp (Uncurry_Proc mode u_info P' R' prf) = 
    debug (i []) ''Uncurry_Proc'' (do {
      dpp' \<leftarrow> uncurry_proc_both I mode u_info P' R' dpp
        <+? (\<lambda>s. i +@+ '': error when applying the uncurrying processor on the DP problem''
          +#+ shows_nl +@+ shows_dpp I dpp +@+ shows_nl +@+ s);
      check_dp_termination_proof assms (i +@+ shows_string ''.1'') I dpp' prf
        <+? (\<lambda>s. i +@+ '': error below the uncurrying processor'' +#+ shows_nl
          +@+ indent s)
    })"
| "check_dp_termination_proof assms i I dpp (Size_Change_Subterm_Proc graphs) =
     debug (i []) ''Size_Change_Subterm_Proc'' (sct_subterm_proc I graphs dpp)"
| "check_dp_termination_proof assms i I dpp (Size_Change_Redpair_Proc redp U_opt graphs) =
     debug (i []) ''Size_Change_Redpair_Proc'' (sct_ur_af_proc I (get_redtriple redp) graphs U_opt dpp)"
| "check_dp_termination_proof assms i I dpp (Fcc_Proc f fcs Pb' R' prf) = debug (i []) ''Fcc_Proc'' (do {
    dpp' \<leftarrow> fcc_proc I f fcs Pb' R' dpp
      <+? (\<lambda>s. i +@+
        '': error when applying the flat context closure processor on the DP problem'' +#+
        shows_nl +@+ shows_dpp I dpp +@+ shows_nl +@+ s);
    check_dp_termination_proof assms (i +@+ shows_string ''.1'') I dpp' prf
      <+? (\<lambda>s. i +@+ '': error below the flat context closure processor'' +#+ shows_nl
        +@+ indent s)
  })"
| "check_dp_termination_proof assms i I dpp (Split_Proc Prem Rrem prf1 prf2) = debug (i []) ''Split_Proc'' (case get_fcc_option prf1 of 
          Some (f,fcs,Pb',Rb',prf1') \<Rightarrow> debug (i []) ''Split_ProcFcc'' (do {
            (dpp1,dpp2) \<leftarrow> fcc_split_proc I f fcs Pb' Rb' Prem Rrem dpp;
            check_dp_termination_proof assms (i +@+ shows_string ''.1.1'') I dpp1 prf1'
              <+? (\<lambda>s. i +@+ '': error below the split and fcc processor'' +#+ shows_nl
                +@+ indent s);
            check_dp_termination_proof assms (i +@+ shows_string ''.2'') I dpp2 prf2
              <+? (\<lambda>s. i +@+ '': error below the split processor'' +#+ shows_nl
                 +@+ indent s)
          })
       | None \<Rightarrow> do {
    let (dpp1,dpp2) = split_proc I dpp Prem Rrem;
    check_dp_termination_proof assms (i +@+ shows_string ''.1'') I dpp1 prf1
      <+? (\<lambda>s. i +@+ '': error below the split processor'' +#+ shows_nl
        +@+ indent s);
    check_dp_termination_proof assms (i +@+ shows_string ''.2'') I dpp2 prf2
      <+? (\<lambda>s. i +@+ '': error below the split processor'' +#+ shows_nl
        +@+ indent s)
  })" 
| "check_dp_termination_proof assms i I dpp (Semlab_Proc sli lP lQ lR prf) = 
    debug (i []) ''Semlab_Proc'' (do {
      dpp' \<leftarrow> semlab_fin_proc I sli lP lQ lR dpp
        <+? (\<lambda>s. i +@+ '': error when applying the semlab processor on the DP problem''
          +#+ shows_nl +@+ shows_dpp I dpp +@+ shows_nl +@+ s);
      check_dp_termination_proof assms (i +@+ shows_string ''.1'') I dpp' prf
        <+? (\<lambda>s. i +@+ '': error below the semlab processor'' +#+ shows_nl
          +@+ indent s)
    })"
| "check_dp_termination_proof assms i I dpp (Switch_Innermost_Proc joins prf) = 
    debug (i []) ''Switch_Innermost_Proc'' (do {
      dpp' \<leftarrow> switch_innermost_proc I joins dpp
        <+? (\<lambda>s. i +@+ '': error when applying the processor to switch to innermost on the DP problem''
          +#+ shows_nl +@+ shows_dpp I dpp +@+ shows_nl +@+ s);
      check_dp_termination_proof assms (i +@+ shows_string ''.1'') I dpp' prf
        <+? (\<lambda>s. i +@+ '': error below the switch to innermost processor'' +#+ shows_nl
          +@+ indent s)
    })"
| "check_dp_termination_proof assms i I dpp (Assume_Finite dpp' dps) =
    debug (i []) ''Finiteness Assumption or Unknown Proof'' (
      if assms
        then (do {
          check_dpp_subsumes I dpp' dpp
            <+? (\<lambda>s. i +@+ '': error in finiteness assumption or unknown proof'' +#+ shows_nl +@+ s +@+ shows_nl);
          check_allm_index (\<lambda>(dpp', p) j. check_dp_termination_proof assms (i +@+ ''.'' +#+ shows (Suc j)) I (mk_dpp I dpp') p) dps
            <+? (\<lambda>s. i +@+ '': error below unknown proof'' +#+ shows_nl
                 +@+ indent s)
        }) else error (i +@+ '': the proof contains a finiteness assumption or unknown proof'' +#+ shows_nl)
    )"
| "check_dp_termination_proof assms i I dpp (Rewriting_Proc U_opt st st' st'' lr p prf) = debug (i []) ''Rewriting_Proc'' (do {
         dpp' \<leftarrow> rewriting_proc I U_opt st st' st'' lr p dpp 
           <+? (\<lambda>s. i +@+ '': error when applying the rewriting processor to rewrite the pair ''
             +#+ shows_nl +@+ shows_rule st +@+ shows_nl +@+
             '' to the pair '' +#+ shows_nl +@+ shows_rule st'' +@+ shows_nl +@+ s);
         check_dp_termination_proof assms (i +@+ shows_string ''.1'') I dpp' prf
           <+? (\<lambda>s. i +@+ '': error below the rewriting processor'' +#+ shows_nl
             +@+ indent s)
   })" 
| "check_dp_termination_proof assms i I dpp (Narrowing_Proc st p sts prf) = debug (i []) ''Narrowing_Proc'' (do {
         dpp' \<leftarrow> narrowing_proc I st p sts dpp 
           <+? (\<lambda>s. i +@+ '': error when applying the narrowing processor to narrow the pair ''
             +#+ shows_nl +@+ shows_rule st +@+ shows_nl +@+
             '' to the pairs '' +#+ shows_nl +@+ shows_trs sts +@+ shows_nl +@+ s);
         check_dp_termination_proof assms (i +@+ shows_string ''.1'') I dpp' prf
           <+? (\<lambda>s. i +@+ '': error below the narrowing processor'' +#+ shows_nl
             +@+ indent s)
   })" 
| "check_dp_termination_proof assms i I dpp (Instantiation_Proc st sts prf) = debug (i []) ''Instantiation_Proc'' (do {
         dpp' \<leftarrow> instantiation_proc I st sts dpp 
           <+? (\<lambda>s. i +@+ '': error when applying the instantiation processor to instantiate the pair ''
             +#+ shows_nl +@+ shows_rule st +@+ shows_nl +@+
             '' to the pairs '' +#+ shows_nl +@+ shows_trs sts +@+ shows_nl +@+ s);
         check_dp_termination_proof assms (i +@+ shows_string ''.1'') I dpp' prf
           <+? (\<lambda>s. i +@+ '': error below the instantiation processor'' +#+ shows_nl
             +@+ indent s)
   })" 
| "check_dp_termination_proof assms i I dpp (Forward_Instantiation_Proc st sts u_opt prf) = debug (i []) ''Forward_Instantiation_Proc'' (do {
         dpp' \<leftarrow> forward_instantiation_proc I st sts u_opt dpp 
           <+? (\<lambda>s. i +@+ '': error when applying the forward_instantiation processor to instantiate the pair ''
             +#+ shows_nl +@+ shows_rule st +@+ shows_nl +@+
             '' to the pairs '' +#+ shows_nl +@+ shows_trs sts +@+ shows_nl +@+ s);
         check_dp_termination_proof assms (i +@+ shows_string ''.1'') I dpp' prf
           <+? (\<lambda>s. i +@+ '': error below the forward instantiation processor'' +#+ shows_nl
             +@+ indent s)
   })" 
| "check_dp_termination_proof assms i I dpp (Unlab_Proc P R prf) = debug (i []) ''Unlab_Proc'' (error (shows ''unlabeling processor not supported''))"
| "check_dp_termination_proof assms i I dpp (General_Redpair_Proc rp ps pb prof prfs) =
    debug (i []) ''General_Redpair_ProcProc'' (do {
      let n = length prfs;
      check (n > 0) (shows ''at least one subproof is required'');
      let merge = (n = 1);
      dpps \<leftarrow> conditional_general_reduction_pair_proc I (get_non_inf_order rp) ps pb prof merge dpp
        <+? (\<lambda>s. i +@+ '': error when applying the generic reduction pair processor to the DP problem''
          +#+ shows_nl +@+ shows_dpp I dpp +@+ shows_nl +@+ s);
      check_dp_termination_proof assms (i +@+ shows_string ''.1'') I (dpps ! 0) (prfs ! 0)
        <+? (\<lambda>s. i +@+ '': error below the generic reduction pair processor'' +#+ shows_nl
          +@+ indent s);
      if merge then succeed else (
        check_dp_termination_proof assms (i +@+ shows_string ''.2'') I (dpps ! 1) (prfs ! 1)
          <+? (\<lambda>s. i +@+ '': error below the generic reduction pair processor'' +#+ shows_nl
            +@+ indent s))
    })"
| "check_dp_termination_proof assms i I dpp (Complex_Constant_Removal_Proc p prf) =
    debug (i []) ''Complex_Constant_Removal_Proc'' (do {
      dpp' \<leftarrow> complex_constant_removal_proc I p dpp
        <+? (\<lambda>s. i +@+ '': error when applying the complex constant removal processor to the DP problem''
          +#+ shows_nl +@+ shows_dpp I dpp
          +@+ shows_nl +@+ s);
      check_dp_termination_proof assms (i +@+ shows_string ''.1'') I dpp' prf
        <+? (\<lambda>s. i +@+ '': error below the complex constant removal processor'' +#+ shows_nl
          +@+ indent s)
    })"
  by pat_completeness auto

termination
proof (rule, unfold update_error_return)
  fix a i I dpp
    and edpts :: "(('f::{show,key}, 'l::{show,key}, string) dp_termination_proof option
      \<times> ('f, 'l, string ) trsLL) list" 
    and pdpps pdpp i' prof dpp' j
  assume proc: "dep_graph_proc I dpp edpts = Inr pdpps"
    and mem: "pdpp \<in> set pdpps" and pdpp: "(prof, dpp') = pdpp"
  from proc mem [unfolded pdpp [symmetric]]
    have "Some prof \<in> set (map fst edpts)" unfolding dep_graph_proc_def by force
  then obtain P where mem: "(Some prof, P) \<in> set edpts" by auto
  let ?m = "prod_size (option_size size)
              (list_size
                (prod_size (Term.term.term_size (lab_size (\<lambda>x. 0) (\<lambda>x. 0)) (\<lambda>x. 0))
                  (Term.term.term_size (lab_size (\<lambda>x. 0) (\<lambda>x. 0)) (\<lambda>x. 0))))"
  have "size prof < ?m (Some prof, P)" by auto
  also have "... \<le> list_size ?m edpts" using mem by (induct edpts, auto)
  finally have "size prof < size (Dep_Graph_Proc edpts)" by simp
  thus "((a, i +@+ ''.'' +#+ shows (Suc j), I, dpp', prof), a, i, I, dpp, Dep_Graph_Proc edpts)
    \<in> measures [\<lambda>(a, i, I, dpp, prof). size prof]" by auto
next
  fix assms i I dpp Prem Rrem prf1 prf2 a x y xa ya xb yb xc yc yd xd ye
  assume ass1: "get_fcc_option prf1 = Some a" and ass2: "(x, y) = a" "(xa, ya) = y"
         "(xb, yb) = ya" "(xc, yc) = yb"
         "fcc_split_proc I x xa xb xc Prem Rrem dpp = Inr yd" "(xd, ye) = yd"
  show "((assms, i +@+ shows_string ''.1.1'', I, xd, yc), assms, i, I,
            dpp, Split_Proc Prem Rrem prf1 prf2)
           \<in> measures [\<lambda>(a, i, I, dpp, prof). size prof]" 
  proof -
    from ass1 obtain f fcs Pb' Rb' prf1a 
      where prf1: "prf1 = Fcc_Proc f fcs Pb' Rb' prf1a" by (cases prf1, force+)
    from ass1 ass2 
    show ?thesis unfolding prf1 by auto
  qed
next
  fix assms i I dpp rp ps pb c prfs x y xa ya
  show  "x = length prfs \<Longrightarrow>
       check (0 < x) (shows ''at least one subproof is required'') = Inr y \<Longrightarrow>
       xa = (x = 1) \<Longrightarrow>
       conditional_general_reduction_pair_proc I (get_non_inf_order rp) ps pb c xa dpp = Inr ya \<Longrightarrow>
       ((assms, i +@+ shows_string ''.1'', I, ya ! 0, prfs ! 0), assms, i, I, dpp, General_Redpair_Proc rp ps pb c prfs)
       \<in> measures [\<lambda>(a, i, I, dpp, prof). size prof]"
    by (cases prfs, auto simp: check_def)
next
  fix assms i I dpp rp ps pb c prfs x y xa ya
  show "x = length prfs \<Longrightarrow>
    check (0 < x) (shows ''at least one subproof is required'') = Inr y \<Longrightarrow>
    xa = (x = 1) \<Longrightarrow> \<not> xa \<Longrightarrow> 
    ((assms, i +@+ shows_string ''.2'', I, ya ! 1, prfs ! 1), assms, i, I, dpp, General_Redpair_Proc rp ps pb c prfs)
    \<in> measures [\<lambda>(a, i, I, dpp, prof). size prof]"
    by (cases prfs, simp add: check_def, cases "tl prfs", auto)
next
  fix assms i I dpp dpp' dps y xa ia xb ya xaa
  show "assms \<Longrightarrow>
    check_dpp_subsumes I dpp' dpp = Inr y \<Longrightarrow>
    xa \<in> set dps \<Longrightarrow>
    (xb, ya) = xa \<Longrightarrow>
    ((assms, i +@+ ''.'' +#+ shows (Suc xaa), I, mk_dpp I xb, ya), assms, i, I, dpp, Assume_Finite dpp' dps)
    \<in> measures [\<lambda>(a, i, I, dpp, prof). size prof]"
    by (induct dps) auto
qed auto

type_synonym
  ('f, 'l, 'v) fassm = "('f, 'l, 'v) dppLL list \<times> ('f, 'l, 'v) dppLL"

fun mk_dpp_set where
  "mk_dpp_set (nfs, m, p, pw, q, r, rw) = (nfs, m, set p, set pw, set q, set r, set rw)"

fun finite_dpp_assm :: "('f, 'l, 'v) fassm \<Rightarrow> bool" where
  "finite_dpp_assm (ds, d) = ((\<forall>d\<in>set ds. finite_dpp (mk_dpp_set d))
    \<longrightarrow> finite_dpp (mk_dpp_set d))"

lemma check_dpp_subsumes_sound:
  assumes I: "dpp_spec I"
    and ok: "isOK (check_dpp_subsumes I dpp' dpp)"
    and fin: "finite_dpp (mk_dpp_set dpp')"
  shows "finite_dpp (dpp_ops.dpp I dpp)"
proof -
  obtain nfs m p pw q r rw where
    dpp': "dpp' = (nfs, m, p, pw, q, r, rw)" by (cases dpp') force+

  let ?m   = "dpp_ops.minimal I dpp"
  let ?nfs = "dpp_ops.nfs I dpp"

  let ?P   = "set (dpp_ops.P I dpp)"
  let ?Pw  = "set (dpp_ops.Pw I dpp)"
  let ?Q   = "set (dpp_ops.Q I dpp)"
  let ?R   = "set (dpp_ops.R I dpp)"
  let ?Rw  = "set (dpp_ops.Rw I dpp)"

  let ?P'  = "set p"
  let ?Pw' = "set pw"
  let ?Q'  = "set q"
  let ?R'  = "set r"
  let ?Rw' = "set rw"

  from check_dpp_subsumesE [OF I ok [unfolded dpp']]
    have flags: "?m = m" "?nfs = nfs"
    and 1: "?P \<subseteq> ?P'"
       "?P \<union> ?Pw \<subseteq> ?P' \<union> ?Pw'"
       "NF_terms ?Q' = NF_terms ?Q"
       "?R \<subseteq> ?R'"
       "?R \<union> ?Rw = ?R' \<union> ?Rw'" 
       by blast+
  let ?dppg = "\<lambda> nfs m. (nfs, m, ?P', ?Pw', ?Q', ?R', ?Rw')"
  show ?thesis
    unfolding dpp_spec.dpp_sound [OF I]
  proof (rule finite_dpp_mono [OF _ 1], rule ccontr)
    assume "\<not> finite_dpp (?dppg ?nfs ?m)"
    then obtain s t \<sigma> where "min_ichain (?dppg ?nfs ?m) s t \<sigma>"
      unfolding finite_dpp_def by auto
    hence "min_ichain (?dppg ?nfs ?m) s t \<sigma>" by auto
    hence "\<not> finite_dpp (?dppg ?nfs ?m)"
      unfolding finite_dpp_def by auto
    with fin [unfolded dpp'] show False
      unfolding flags by simp
  qed
qed

lemma isOK_check_dp_termination_proof_Assume_Finite_False [simp]:
  "isOK (check_dp_termination_proof False i I dpp (Assume_Finite dpp' dps)) = False" by simp

fun dp_assms :: "bool \<Rightarrow> ('f, 'l, 'v) dp_termination_proof \<Rightarrow> ('f, 'l, 'v) fassm list" where
  "dp_assms a P_is_Empty = []"
| "dp_assms a (Subterm_Criterion_Proc _ _ _ p) = dp_assms a p"
| "dp_assms a (Redpair_Proc _ _ p) = dp_assms a p"
| "dp_assms a (Redpair_UR_Proc _ _ _ p) = dp_assms a p"
| "dp_assms a (Usable_Rules_Proc _ p) = dp_assms a p"
| "dp_assms a (Dep_Graph_Proc ps) =
    concat (concat (map (map (dp_assms a) \<circ> option_to_list \<circ> fst) ps))"
| "dp_assms a (Mono_Redpair_Proc _ _ _ p) = dp_assms a p"
| "dp_assms a (Mono_Redpair_UR_Proc _ _ _ _ p) = dp_assms a p"
| "dp_assms a (Size_Change_Subterm_Proc _) = []"
| "dp_assms a (Size_Change_Redpair_Proc _ _ _) = []"
| "dp_assms a (Uncurry_Proc _ _ _ _ p) = dp_assms a p"
| "dp_assms a (Fcc_Proc _ _ _ _ p) = dp_assms a p"
| "dp_assms a (Split_Proc _ _ p q) = dp_assms a p @ dp_assms a q"
| "dp_assms a (Semlab_Proc _ _ _ _ p) = dp_assms a p"
| "dp_assms a (Switch_Innermost_Proc _ p) = dp_assms a p"
| "dp_assms a (Rewriting_Proc _ _ _ _ _ _ p) = dp_assms a p"
| "dp_assms a (Narrowing_Proc _ _ _ p) = dp_assms a p"
| "dp_assms a (Instantiation_Proc _ _ p) = dp_assms a p"
| "dp_assms a (Forward_Instantiation_Proc _ _ _ p) = dp_assms a p"
| "dp_assms a (Assume_Finite d dps) = (if a then (map fst dps, d) # concat (map (dp_assms a \<circ> snd) dps) else [])"
| "dp_assms a (Unlab_Proc _ _ p) = dp_assms a p"
| "dp_assms a (Q_Reduction_Proc _ p) = dp_assms a p"
| "dp_assms a (General_Redpair_Proc _ _ _ _ ps) = concat (map (dp_assms a) ps)"
| "dp_assms a (Complex_Constant_Removal_Proc _ p) = dp_assms a p"

lemma dp_assms_False [simp]: "dp_assms False prf = []"
  by (induct False "prf" rule: dp_assms.induct) simp_all

lemma check_dp_termination_proof_with_dp_assms:
  assumes I: "dpp_spec I"
    and fin: "\<forall>assm\<in>set (dp_assms a prf). finite_dpp_assm assm" (is "?P a prf")
    and ok: "isOK (check_dp_termination_proof a i I dpp prf)"
  shows "finite_dpp (dpp_ops.dpp I dpp)"
using ok I fin
proof (induct a i I dpp "prf" rule: check_dp_termination_proof.induct)
  case (1 a i I dpp)
  note IH = this
  interpret dpp_spec I by fact
  from IH(1) show ?case by (auto simp:
    finite_dpp_def ichain.simps)
next
  case (2 a i I dpp p rseq rP prof)
  note IH = this
  interpret dpp_spec I by fact
  let ?Pr = "ceta_list_diff (dpp_ops.pairs I dpp) rP"
  let ?i = "i +@+ shows_string ''.1''"
  from IH(2)
    obtain dpp' where proc: "subterm_criterion_proc I p rseq ?Pr dpp = return dpp'"
    and rec: "isOK (check_dp_termination_proof a ?i I dpp' prof)"
    by (auto simp: Let_def)
  from IH have fin: "?P a prof" by simp
  show ?case
    by (rule sound_proc_impl[OF subterm_criterion_proc[OF IH(3)] proc],
      rule IH(1)[OF HOL.refl HOL.refl _ rec IH(3) fin], insert proc, simp)
next
  case (3 a i I dpp rpc rP prof)
  note IH = this
  interpret dpp_spec I by fact
  let ?Pr = "ceta_list_diff (dpp_ops.pairs I dpp) rP"
  let ?i = "i +@+ shows_string ''.1''"
  from IH have fin: "?P a prof" by simp
  show ?case 
  proof (cases rpc)
    case (Inr rp)
    with IH(2)
    obtain dpp' where proc: "generic_ur_af_redtriple_proc I (get_redtriple rp) None ?Pr dpp = return dpp'"
      and rec: "isOK (check_dp_termination_proof a ?i I dpp' prof)"
      by (auto simp: Let_def)
    from IH(1)[OF HOL.refl HOL.refl HOL.refl _ rec IH(3) fin, unfolded Inr update_error_return sum.simps, OF proc]
    have fin: "finite_dpp (dpp_ops.dpp I dpp')" .
    show ?thesis      
      by (rule sound_proc_impl[OF generic_ur_af_redtriple_proc[OF IH(3) get_redtriple] _ fin], 
        rule proc)
  next
    case (Inl rrp)
    with IH(2)
    obtain dpp' where proc: "generic_ur_af_root_redtriple_proc I (get_root_redtriple rrp) None ?Pr dpp = return dpp'"
      and rec: "isOK (check_dp_termination_proof a ?i I dpp' prof)"
      by (auto simp: Let_def)
    from IH(1)[OF HOL.refl HOL.refl HOL.refl _ rec IH(3) fin, unfolded Inl update_error_return sum.simps, OF proc] 
    have fin: "finite_dpp (dpp_ops.dpp I dpp')" .
    show ?thesis
      by (rule sound_proc_impl[OF generic_ur_af_root_redtriple_proc[OF IH(3) get_root_redtriple] _ fin], rule proc)
  qed
next
  case (4 a i I dpp ur prof)
  note IH = this
  interpret dpp_spec I by fact
  let ?i = "i +@+ shows_string ''.1''"
  from IH(2) obtain dpp'
    where proc: "usable_rules_proc I ur dpp = return dpp'"
    and rec: "isOK (check_dp_termination_proof a ?i I dpp' prof)"
    by (auto simp: Let_def)
  from IH have fin: "?P a prof" by simp
  show ?case
    by (rule sound_proc_impl[OF usable_rules_proc, OF _ proc 4(1)[OF _ rec _ fin]], insert proc 4(3), auto)
next
  case (5 a i I dpp q prof)
  note IH = this
  interpret dpp_spec I by fact
  let ?i = "i +@+ shows_string ''.1''"
  from IH(2)
    obtain dpp' where proc: "q_reduction_proc I q dpp = return dpp'"
      and rec: "isOK (check_dp_termination_proof a ?i I dpp' prof)"
      by (auto simp: Let_def)
  from IH have fin: "?P a prof" by simp
  show ?case      
    by (rule sound_proc_impl[OF q_reduction_proc[OF IH(3)], OF proc IH(1), 
      unfolded update_error_return, OF proc rec IH(3) fin]) 
next
  case (6 a i I dpp rp rP rR prof)
  note IH = this
  interpret dpp_spec I by fact
  let ?Pr = "ceta_list_diff (dpp_ops.pairs I dpp) rP"
  let ?Rr = "ceta_list_diff (dpp_ops.rules I dpp) rR"
  let ?i = "i +@+ shows_string ''.1''"
  from IH(2)
    obtain dpp' where proc: "mono_redpair_proc I (get_redtriple rp) ?Pr ?Rr dpp = return dpp'"
    and rec: "isOK (check_dp_termination_proof a ?i I dpp' prof)"
    by (auto simp: Let_def)
  from IH have fin: "?P a prof" by simp
  show ?case      
    by (rule sound_proc_impl[OF mono_redpair_proc[OF get_redtriple],
      OF proc IH(1)[OF HOL.refl HOL.refl HOL.refl HOL.refl _ rec IH(3) fin]],
      unfold update_error_return, rule proc)
next
  case (7 a i I dpp edpts)
  note IH = this
  interpret dpp_spec I by fact
  from IH(2) obtain pdpps where proc: "dep_graph_proc I dpp edpts = return pdpps" 
    and rec: "\<And>j. j < length pdpps \<Longrightarrow>
      isOK ((\<lambda>(prof, dpp') j.
        check_dp_termination_proof a (i +@+ ''.'' +#+ shows (Suc j)) I dpp' prof) (pdpps ! j) j)"
    by auto
  show ?case
  proof (rule dep_graph_proc[OF IH(3) proc])
    fix p dpp'
    assume mem: "(p, dpp') \<in> set pdpps"
    from this[unfolded set_conv_nth]
      obtain j where j: "j < length pdpps" and pair: "pdpps ! j = (p, dpp')" by auto
    from rec[OF j] have rec: "isOK (check_dp_termination_proof a (i +@+ ''.'' +#+ shows (Suc j)) I dpp' p)"
      unfolding pair by simp
    from proc[unfolded dep_graph_proc_def, simplified]
      have pdpps: "pdpps =
      map (\<lambda>aP. (the (fst aP), dpp_ops.intersect_pairs I dpp (snd aP)))
          [aP\<leftarrow>edpts . \<exists>y. fst aP = Some y]" by simp
    from mem and IH(4) and proc
      have fin: "?P a p" by (simp add: dep_graph_proc_def) force
    show "finite_dpp (dpp_ops.dpp I dpp')" 
      by (rule IH(1)[unfolded update_error_return, OF proc mem HOL.refl rec IH(3) fin])
  qed
next
  case (8 a i I dpp rpc rP ur prof)
  note IH = this
  interpret dpp_spec I by fact
  let ?Pr = "ceta_list_diff (dpp_ops.pairs I dpp) rP"
  let ?i = "i +@+ shows_string ''.1''"
  from IH have fin: "?P a prof" by simp
  show ?case 
  proof (cases rpc)
    case (Inr rp)
    with IH(2) obtain dpp'
      where proc: "generic_ur_af_redtriple_proc I (get_redtriple rp) (Some ur) ?Pr dpp = return dpp'"
      and rec: "isOK (check_dp_termination_proof a ?i I dpp' prof)"
      by (auto simp: Let_def)
    show ?thesis      
      by (rule sound_proc_impl[OF generic_ur_af_redtriple_proc[OF IH(3) get_redtriple],
        OF proc IH(1)[OF HOL.refl HOL.refl HOL.refl _ rec IH(3) fin]],
        unfold Inr update_error_return sum.simps, rule proc)
  next
    case (Inl rrp)
    with IH(2) obtain dpp'
      where proc: "generic_ur_af_root_redtriple_proc I (get_root_redtriple rrp) (Some ur) ?Pr dpp = return dpp'"
      and rec: "isOK (check_dp_termination_proof a ?i I dpp' prof)" 
      by (auto simp: Let_def)
    show ?thesis      
      by (rule sound_proc_impl[OF generic_ur_af_root_redtriple_proc[OF IH(3) get_root_redtriple],
        OF proc IH(1)[OF HOL.refl HOL.refl HOL.refl _ rec IH(3) fin]],
        unfold Inl update_error_return sum.simps, rule proc)
  qed
next
  case (9 a i I dpp rp rP rR ur prof)
  note IH = this
  interpret dpp_spec I by fact
  let ?Pr = "ceta_list_diff (dpp_ops.pairs I dpp) rP"
  let ?Rr = "ceta_list_diff (dpp_ops.rules I dpp) rR"
  let ?i = "i +@+ shows_string ''.1''"
  from IH(2) obtain dpp'
    where proc: "mono_ur_redpair_proc I (get_redtriple rp) ?Pr ?Rr ur dpp = return dpp'"
    and rec: "isOK (check_dp_termination_proof a ?i I dpp' prof)"
    by (auto simp: Let_def)
  from IH have fin: "?P a prof" by simp
  show ?case
    by (rule sound_proc_impl[OF mono_ur_redpair_proc[OF get_redtriple],
      OF proc IH(1)[OF HOL.refl HOL.refl HOL.refl HOL.refl _ rec IH(3) fin]],
      unfold update_error_return, rule proc)
next
  case (10 a i I dpp mode u_info P' R' prof)
  note IH = this
  interpret dpp_spec I by fact
  let ?i = "i +@+ shows_string ''.1''"
  from IH(2) obtain
    dpp' where proc: "uncurry_proc_both I mode u_info P' R'  dpp = return dpp'"
    and rec: "isOK (check_dp_termination_proof a ?i I dpp' prof)"
    by (auto simp: Let_def)
  from IH have fin: "?P a prof" by simp
  show ?case      
    by (rule sound_proc_impl[OF uncurry_proc_both[OF IH(3)], OF proc IH(1)],
      unfold update_error_return, rule proc, rule rec, rule IH(3), rule fin)
next
  case (11 a i I dpp graphs)
  note IH = this
  from IH(1) have proc: "isOK (sct_subterm_proc I graphs dpp)" by auto
  show ?case 
    by (rule sct_subterm_proc_sound[OF `dpp_spec I` proc])
next
  case (12 a i I dpp rp u_opt graphs)
  note IH = this
  interpret dpp_spec I by fact
  from IH(1) have proc: "isOK (sct_ur_af_proc I (get_redtriple rp) graphs u_opt dpp)" by auto
  show ?case
    by (rule sct_ur_af_proc[OF IH(2) get_redtriple, OF proc])
next
  case (13 a i I dpp f fcs Pb' R' prof)
  note IH = this
  interpret dpp_spec I by fact
  let ?i = "i +@+ shows_string ''.1''"
  from IH(2) obtain dpp'
    where proc: "fcc_proc I f fcs Pb' R' dpp = return dpp'"
    and rec: "isOK (check_dp_termination_proof a ?i I dpp' prof)" by auto
  from IH have fin: "?P a prof" by simp
  from IH(1)[OF _ rec IH(3) fin, simplified, OF proc] have "finite_dpp (dpp_ops.dpp I dpp')"
    by (simp add: dpp_spec_sound)
  from sound_proc_impl[OF fcc_proc, OF proc this] show ?case .
next
  case (14 a i I dpp P' R' prof1 prof2)
  note IH = this
  interpret dpp_spec I by fact
  let ?i = "\<lambda>s. i +@+ shows_string s"
  let ?i1 = "?i ''.1''"
  let ?i11 = "?i ''.1.1''"
  let ?i2 = "?i ''.2''"
  note res = IH(5)[unfolded check_dp_termination_proof.simps debug.simps]
  note ass = IH(7)[unfolded dp_assms.simps]
  show ?case
  proof (cases "get_fcc_option prof1")
    case None
    obtain dpp1 dpp2 where s: "split_proc I dpp P' R' = (dpp1, dpp2)" by force
    note ss = s[symmetric]
    note res = res[unfolded None s option.simps Let_def split]
    from res have ok1: "isOK (check_dp_termination_proof a ?i1 I dpp1 prof1)"
    and ok2: "isOK (check_dp_termination_proof a ?i2 I dpp2 prof2)" by auto
    from ass have fin1: "?P a prof1" and fin2: "?P a prof2" by simp_all
    from IH(1)[OF None ss refl ok1 IH(6) fin1]
    have fin1': "finite_dpp (dpp_ops.dpp I dpp1)" .
    from IH(2)[OF None ss refl _ ok2 IH(6) fin2] ok1
    have fin2': "finite_dpp (dpp_ops.dpp I dpp2)" by auto
    show ?thesis
      by (rule split_proc[OF s fin1' fin2'])
  next
    case (Some fcco)
    from Some obtain f fcs Pb' Rb' prf1a where prof1: "prof1 = Fcc_Proc f fcs Pb' Rb' prf1a" by (cases prof1, force+)
    hence fcco: "get_fcc_option prof1 = Some (f,fcs,Pb',Rb',prf1a)" by simp
    note res = res[unfolded option.simps fcco Let_def split]
    let ?call = "fcc_split_proc I f fcs Pb' Rb' P' R' dpp"
    from res obtain dpp1 dpp2 where proc: "?call = Inr (dpp1,dpp2)" 
      by (cases ?call, auto)
    note res = res[unfolded proc, simplified]
    from res have ok1: "isOK (check_dp_termination_proof a ?i11 I dpp1 prf1a)"
      and ok2: "isOK (check_dp_termination_proof a ?i2 I dpp2 prof2)" by auto
    from ass have fin1: "?P a prf1a" and fin2: "?P a prof2" 
      unfolding prof1 by simp_all
    from IH(3)[OF fcco refl refl refl refl proc refl ok1 IH(6) fin1]
    have fin1': "finite_dpp (dpp_ops.dpp I dpp1)" .
    from IH(4)[OF fcco refl refl refl refl proc refl _ ok2 IH(6) fin2] ok1
    have fin2': "finite_dpp (dpp_ops.dpp I dpp2)" by auto
    show ?thesis
      by (rule fcc_split_proc[OF IH(6) proc fin1' fin2'])
  qed
next
  case (15 a i I dpp sli lP lQ lR prof)
  note IH = this
  interpret dpp_spec I by fact
  let ?i = "i +@+ shows_string ''.1''"
  from IH(2) obtain dpp'
    where proc: "semlab_fin_proc I sli lP lQ lR dpp = return dpp'"
    and rec: "isOK (check_dp_termination_proof a ?i I dpp' prof)"
    by (auto simp: Let_def)
  from IH have fin: "?P a prof" by simp
  show ?case
    by (rule sound_proc_impl[OF semlab_fin_proc[OF IH(3)], OF proc IH(1)],
      unfold update_error_return, rule proc, rule rec, rule IH(3), rule fin)
next
  case (16 a i I dpp joins prof)
  note IH = this
  interpret dpp_spec I by fact
  let ?i = "i +@+ shows_string ''.1''"
  from IH(2) obtain dpp'
    where proc: "switch_innermost_proc I joins dpp = return dpp'"
    and rec: "isOK (check_dp_termination_proof a ?i I dpp' prof)"
    by (auto simp: Let_def)
  from IH have fin: "?P a prof" by simp
  show ?case    
    by (rule sound_proc_impl[OF switch_innermost_proc[OF IH(3)] _ IH(1)], rule proc, unfold update_error_return, rule proc, rule rec, rule IH(3), rule fin)
next
  case (17 a i I dpp dpp' dps)
  interpret dpp_spec I by fact
  show ?case
  proof (cases a)
    assume "\<not> a"
    then show ?thesis
      using `isOK (check_dp_termination_proof a i I dpp (Assume_Finite dpp' dps))`
      by simp
  next
    assume "a"
    have ok: "isOK (check_dp_termination_proof a i I dpp (Assume_Finite dpp' dps))" by fact
    then have subsumes: "isOK (check_dpp_subsumes I dpp' dpp)" by simp
    have "(try check_dpp_subsumes I dpp' dpp catch (\<lambda>x. Inl (i +@+
      '': error in finiteness assumption or unknown proof'' +#+ shows_nl +@+ x +@+
      shows_nl))) = Inr ()" using subsumes by auto
    note IH = 17(1) [OF `a` this _ _ _ `dpp_spec I`]
    show ?thesis
    proof (rule check_dpp_subsumes_sound [OF `dpp_spec I` subsumes])
      have fin: "\<forall>a\<in>set (dp_assms a (Assume_Finite dpp' dps)). finite_dpp_assm a" by fact
      then have *: "(\<And>d. d \<in> set (map fst dps) \<Longrightarrow> finite_dpp (mk_dpp_set d))
        \<Longrightarrow> finite_dpp (mk_dpp_set dpp')" by (simp add: `a`)
      show "finite_dpp (mk_dpp_set dpp')"
      proof (rule *)
        fix d assume "d \<in> set (map fst dps)"
        then obtain p where "(d, p) \<in> set dps" by auto
        then obtain j where j: "j < length dps" "dps ! j = (d, p)" by (induct dps) force+
        have "finite_dpp (dpp_ops.dpp I (mk_dpp I d))"
        proof (rule IH)
          show "(d, p) \<in> set dps" by fact
        next
          show "(d, p) = (d, p)" by (rule refl)
        next
          from j and ok
            show "isOK (check_dp_termination_proof a (i +@+ ''.'' +#+ shows (Suc j)) I (mk_dpp I d) p)"
            by auto
        next
          from fin show "\<forall>a\<in>set (dp_assms a p). finite_dpp_assm a"
            using `(d, p) \<in> set dps` and `a` by auto
        qed
        then show "finite_dpp (mk_dpp_set d)" by (cases d) simp
      qed
    qed
  qed
next
  case (18 a i I dpp U st st' st'' lr p prof)
  note IH = this
  interpret dpp_spec I by fact
  let ?i = "i +@+ shows_string ''.1''"
  from IH(2) obtain dpp'
    where proc: "rewriting_proc I U st st' st'' lr p dpp = return dpp'"
    and rec: "isOK (check_dp_termination_proof a ?i I dpp' prof)"
    by (auto simp: Let_def)
  from IH have fin: "?P a prof" by simp
  show ?case    
    by (rule sound_proc_impl[OF rewriting_proc[OF IH(3)] _ IH(1)], rule proc, unfold update_error_return, rule proc, rule rec, rule IH(3), rule fin)  
next
  case (19 a i I dpp st p sts prof)
  note IH = this
  interpret dpp_spec I by fact
  let ?i = "i +@+ shows_string ''.1''"
  from IH(2) obtain dpp'
    where proc: "narrowing_proc I st p sts dpp = return dpp'"
    and rec: "isOK (check_dp_termination_proof a ?i I dpp' prof)"
    by (auto simp: Let_def)
  from IH have fin: "?P a prof" by simp
  show ?case    
    by (rule sound_proc_impl[OF narrowing_proc[OF IH(3)] _ IH(1)], rule proc, unfold update_error_return, rule proc, rule rec, rule IH(3), rule fin)  
next
  case (20 a i I dpp st sts prof)
  note IH = this
  interpret dpp_spec I by fact
  let ?i = "i +@+ shows_string ''.1''"
  from IH(2) obtain dpp'
    where proc: "instantiation_proc I st sts dpp = return dpp'"
    and rec: "isOK (check_dp_termination_proof a ?i I dpp' prof)"
    by (auto simp: Let_def)
  from IH have fin: "?P a prof" by simp
  show ?case    
    by (rule sound_proc_impl[OF instantiation_proc[OF IH(3)] _ IH(1)], rule proc, unfold update_error_return, rule proc, rule rec, rule IH(3), rule fin)  
next
  case (21 a i I dpp st sts u_opt prof)
  note IH = this
  interpret dpp_spec I by fact
  let ?i = "i +@+ shows_string ''.1''"
  from IH(2) obtain dpp'
    where proc: "forward_instantiation_proc I st sts u_opt dpp = return dpp'"
    and rec: "isOK (check_dp_termination_proof a ?i I dpp' prof)"
    by (auto simp: Let_def)
  from IH have fin: "?P a prof" by simp
  show ?case    
    by (rule sound_proc_impl[OF forward_instantiation_proc[OF IH(3)] _ IH(1)], rule proc, unfold update_error_return, rule proc, rule rec, rule IH(3), rule fin)
next
  case (22 a i I dpp P R prof)
  thus ?case by simp
next
  case (23 a i I dpp rp Ps Pb prof prfs)
  note IH = this
  interpret dpp_spec I by fact
  let ?i1 = "i +@+ shows_string ''.1''"
  let ?i2 = "i +@+ shows_string ''.2''"
  note prems = IH(3)[simplified, unfolded Let_def check_def, simplified]
  from prems obtain dpps
    where proc: "conditional_general_reduction_pair_proc I (get_non_inf_order rp) Ps Pb prof (length prfs = 1) dpp = return dpps"
      and len: "length prfs \<noteq> 0"
    by auto
  from len obtain p1 ps where prfs: "prfs = p1 # ps" by (cases prfs, auto)
  from conditional_general_reduction_pair_proc_len[OF proc] have len: "length dpps = (if length prfs = 1 then 1 else 2)" .
  from len obtain d1 ds where dpps: "dpps = d1 # ds" by (cases dpps, cases "length prfs = 1", auto)
  from prems proc prfs dpps have ok1: "isOK(check_dp_termination_proof a ?i1 I (dpps ! 0) (prfs ! 0))" by simp
  note IH1 = IH(1)[OF refl _ refl _ ok1 IH(4), unfolded proc prfs]
  from IH(5)[unfolded prfs] have ass1: "?P a p1" by auto
  show ?case
  proof (rule conditional_general_reduction_pair_proc[OF IH(4) get_non_inf_order proc])
    fix d
    assume d: "d \<in> set dpps"
    show "finite_dpp (dpp_ops.dpp I d)"
    proof (cases "d = d1")
      case True
      hence d: "d = (dpps ! 0)" unfolding dpps by simp
      show ?thesis unfolding d
        by (rule IH1, insert proc ass1 prfs, auto simp: check_def)
    next
      case False
      with d dpps have d: "d \<in> set ds" by auto
      with len dpps prfs have dpps: "dpps = [d1,d]" by (cases ps, auto, cases ds, auto)
      with len obtain p2 pps where prfs: "prfs = p1 # p2 # pps" unfolding prfs dpps by (cases ps, auto)
      from dpps have d: "d = dpps ! 1" by simp
      from prems proc prfs dpps have ok2: "isOK(check_dp_termination_proof a ?i2 I (dpps ! 1) (prfs ! 1))" by auto
      from IH(5)[unfolded prfs] have ass2: "?P a p2" by auto
      show ?thesis unfolding d
        by (rule IH(2)[OF refl _ refl _ _ _ ok2 IH(4)], insert proc prfs ok1 ass2 dpps, auto simp: check_def)
    qed
  qed
next
  case (24 a i I dpp cprf prof)
  note IH = this
  interpret dpp_spec I by fact
  let ?i = "i +@+ shows_string ''.1''"
  from IH(2) obtain dpp'
    where proc: "complex_constant_removal_proc I cprf dpp = return dpp'"
    and rec: "isOK (check_dp_termination_proof a ?i I dpp' prof)" by auto
  from IH have fin: "?P a prof" by simp
  from IH(1)[OF _ rec IH(3) fin, simplified, OF proc] have "finite_dpp (dpp_ops.dpp I dpp')"
    by (simp add: dpp_spec_sound)
  from sound_proc_impl[OF complex_constant_removal_proc[OF IH(3)] proc this] show ?case .
qed

(*the old soundness lemma is an instance of the new one*)
lemma check_dp_termination_proof:
  assumes "dpp_spec I"
    and "isOK (check_dp_termination_proof False i I dpp prf)"
  shows "finite_dpp (dpp_ops.dpp I dpp)"
  using check_dp_termination_proof_with_dp_assms[OF assms(1) _ assms(2)] by simp

(* provide a method to eliminate unlab *)
fun update where "update p_to_p (p, info) = (p_to_p p, info)"

definition "no_decr \<equiv> filter (\<lambda>(l, r). \<not> (l \<noteq> r \<and> map_term unlab l = map_term unlab r))"

definition "unlab_of_trs r \<equiv> unlab_rules (no_decr r)"

fun unlab_to_split_proc :: "('f :: key, 'l :: key, 'v :: key) dp_termination_proof \<Rightarrow> ('f, 'l, 'v) dp_termination_proof \<times> (('f,'l,'v)trsLL \<times> ('f,'l,'v)trsLL \<times> ('f,'l,'v) dp_termination_proof) list" where
  "unlab_to_split_proc P_is_Empty = (P_is_Empty,[])"
| "unlab_to_split_proc (Dep_Graph_Proc ps) = (Dep_Graph_Proc
   (map (\<lambda> (po,P). (Option.map (fst o unlab_to_split_proc) po,P)) ps), [])" 
| "unlab_to_split_proc (Subterm_Criterion_Proc a1 a2 a3 p) = update (Subterm_Criterion_Proc a1 a2 a3) (unlab_to_split_proc p)" 
| "unlab_to_split_proc (Redpair_Proc a1 a2 p) = update (Redpair_Proc a1 a2) (unlab_to_split_proc p)"
| "unlab_to_split_proc (Redpair_UR_Proc a1 a2 a3 p) = update (Redpair_UR_Proc a1 a2 a3) (unlab_to_split_proc p)"
| "unlab_to_split_proc (Usable_Rules_Proc a1 p) = update (Usable_Rules_Proc a1) (unlab_to_split_proc p)" 
| "unlab_to_split_proc (Q_Reduction_Proc a1 p) = update (Q_Reduction_Proc a1) (unlab_to_split_proc p)" 
| "unlab_to_split_proc (Mono_Redpair_Proc a1 a2 a3 p) = update (Mono_Redpair_Proc a1 a2 a3) (unlab_to_split_proc p)"
| "unlab_to_split_proc (Mono_Redpair_UR_Proc a1 a2 a3 a4 p) = update (Mono_Redpair_UR_Proc a1 a2 a3 a4) (unlab_to_split_proc p)"
| "unlab_to_split_proc (Size_Change_Subterm_Proc a1) = (Size_Change_Subterm_Proc a1, [])"
| "unlab_to_split_proc (Size_Change_Redpair_Proc a1 a2 a3) = (Size_Change_Redpair_Proc a1 a2 a3, [])"
| "unlab_to_split_proc (Uncurry_Proc a1 a2 a3 a4 p) = update (Uncurry_Proc a1 a2 a3 a4) (unlab_to_split_proc p)"
| "unlab_to_split_proc (Fcc_Proc a1 a2 a3 a4 p) = update (Fcc_Proc a1 a2 a3 a4) (unlab_to_split_proc p)"
| "unlab_to_split_proc (Switch_Innermost_Proc a1 p) = update (Switch_Innermost_Proc a1) (unlab_to_split_proc p)"
| "unlab_to_split_proc (Rewriting_Proc a1 a2 a3 a4 a5 a6 p) = update (Rewriting_Proc a1 a2 a3 a4 a5 a6) (unlab_to_split_proc p)" 
| "unlab_to_split_proc (Narrowing_Proc a1 a2 a3 p) = update (Narrowing_Proc a1 a2 a3) (unlab_to_split_proc p)"
| "unlab_to_split_proc (Instantiation_Proc a1 a2 p) = update (Instantiation_Proc a1 a2) (unlab_to_split_proc p)"
| "unlab_to_split_proc (Forward_Instantiation_Proc a1 a2 a3 p) = update (Forward_Instantiation_Proc a1 a2 a3) (unlab_to_split_proc p)"
| "unlab_to_split_proc (Assume_Finite d dps) = (Assume_Finite d dps, [])"
| "unlab_to_split_proc (Unlab_Proc P R p) = (P_is_Empty,[(P,R,p)])"
| "unlab_to_split_proc (Split_Proc a1 a2 p q) = (Split_Proc a1 a2 (fst (unlab_to_split_proc p)) (fst (unlab_to_split_proc q)), [])" 
| "unlab_to_split_proc (Semlab_Proc a1 lP a2 lR p) = (
     case unlab_to_split_proc p of
        (p,list) \<Rightarrow> let sl = Semlab_Proc a1 lP a2 lR p
        in case list of 
              Nil \<Rightarrow> (sl,list)
            | Cons (P,R,prof) list' \<Rightarrow> 
          (let ulP = unlab_of_trs lP; ulR = unlab_of_trs lR; PR = ceta_list_diff ulP P; RR = ceta_list_diff ulR R
            in (Split_Proc PR RR sl prof,list'))
     )"  
| "unlab_to_split_proc (General_Redpair_Proc a1 a2 a3 a4 ps) = (General_Redpair_Proc a1 a2 a3 a4 ps, [])" (* TODO: implement better routine for General_Redpair_Proc *) 
| "unlab_to_split_proc (Complex_Constant_Removal_Proc a1 p) = update (Complex_Constant_Removal_Proc a1) (unlab_to_split_proc p)"

lemma unlab_to_split_proc_assms: "set (dp_assms a (fst (unlab_to_split_proc prf))) \<subseteq> set (dp_assms a prf)"
proof -
  let ?a = "dp_assms a"
  let ?ap = "\<lambda> p. set (?a p)"
  let ?aps = "\<lambda> pl. ?ap (fst pl) \<union> \<Union>(set (map (\<lambda> (_,_,p). ?ap p) (snd pl)))"
  have "?aps (unlab_to_split_proc prf) \<subseteq> ?ap prf"
  proof (induct "prf" rule: unlab_to_split_proc.induct)
    case 1 thus ?case by simp
  next
    case (2 ps)
    note IH = this
    let ?m = "(\<lambda>(po,y). (Option.map (fst o unlab_to_split_proc) po,y))"
    let ?ps = "map ?m ps"
    let ?f = "(\<lambda>(po,y). \<Union> (?ap ` Option.set po))"
    have "?aps (unlab_to_split_proc (Dep_Graph_Proc ps)) = ?ap (Dep_Graph_Proc ?ps)" by simp
    also have "... = \<Union> (set (map ?f ?ps))" 
      by force 
    also have "... \<subseteq> \<Union> (set (map ?f ps))"
    proof(rule subsetI)
      fix r
      assume "r \<in> \<Union> (set (map ?f ?ps))"
      then obtain x where r: "r \<in> ?f x" and "x \<in> set ?ps" by force+
      then obtain y where x: "x = ?m y" and mem: "y \<in> set ps"  by auto
      obtain y1 y2 where y: "y = (y1,y2)" by force
      note x = x[unfolded y split]
      have "?f x \<subseteq> ?f y" unfolding y x split using IH[OF mem y[symmetric]]
        by (cases y1, auto)
      with mem have "?f x \<subseteq> \<Union> (set (map ?f ps))" by auto
      with r show "r \<in> \<Union> (set (map ?f ps))" by blast
    qed
    also have "... = ?ap (Dep_Graph_Proc ps)" by force
    finally show ?case .
  next
    case (3 a1 a2 a3 p)
    thus ?case by (cases "unlab_to_split_proc p", auto)
  next
    case (4 a1 a2 p)
    thus ?case by (cases "unlab_to_split_proc p", auto)
  next
    case (5 a1 a2 a3 p)
    thus ?case by (cases "unlab_to_split_proc p", auto)
  next
    case (6 a1 p)
    thus ?case by (cases "unlab_to_split_proc p", auto)
  next
    case (7 a1 p)
    thus ?case by (cases "unlab_to_split_proc p", auto)
  next
    case (8 a1 a2 a3 p)
    thus ?case by (cases "unlab_to_split_proc p", auto)
  next
    case (9 a1 a2 a3 a4 p)
    thus ?case by (cases "unlab_to_split_proc p", auto)
  next
    case (10 a1)
    thus ?case by simp
  next
    case (11 a1)
    thus ?case by simp
  next
    case (12 a1 a2 a3 a4 p)
    thus ?case by (cases "unlab_to_split_proc p", auto)
  next
    case (13 a1 a2 a3 a4 p)
    thus ?case by (cases "unlab_to_split_proc p", auto)
  next
    case (14 a1 p)
    thus ?case by (cases "unlab_to_split_proc p", auto)
  next
    case (15 a1 a2 a3 a4 a5 a6 p)
    thus ?case by (cases "unlab_to_split_proc p", auto)
  next
    case (16 a1 a2 a3 p)
    thus ?case by (cases "unlab_to_split_proc p", auto)
  next
    case (17 a1 a2 p)
    thus ?case by (cases "unlab_to_split_proc p", auto)
  next
    case (18 a1 a2 a3 p)
    thus ?case by (cases "unlab_to_split_proc p", auto)
  next
    case (19 a1)
    thus ?case by simp
  next
    case (20 P R p)
    thus ?case by simp
  next
    case (21 a1 a2 p q)
    thus ?case by (cases "unlab_to_split_proc p", cases "unlab_to_split_proc q", auto)
  next
    case (22 a1 lP a2 lR p)
    note IH = this
    obtain p' list where up: "unlab_to_split_proc p = (p',list)" by force
    show ?case
    proof (cases list)
      case Nil
      show ?thesis using IH
        unfolding unlab_to_split_proc.simps up fst_conv snd_conv split Let_def Nil
        by auto
    next
      case (Cons prP list')
      obtain P R p'' where prP: "prP = (P,R,p'')" by (cases prP, auto)
      show ?thesis using IH
        unfolding unlab_to_split_proc.simps up fst_conv snd_conv split Let_def Cons prP list.simps by auto
    qed
  next
    case (23 a1 a2 a3 a4 ps)
    show ?case by simp
  next
    case (24 a1 p)
    thus ?case by (cases "unlab_to_split_proc p", auto)
  qed
  thus ?thesis by auto
qed

hide_const Check_DPP.update

definition check_dp_termination_proof_with_unlab where 
  "check_dp_termination_proof_with_unlab a i I dpp prf \<equiv> check_dp_termination_proof a i I dpp (fst (unlab_to_split_proc prf))"
 
lemma check_dp_termination_proof_with_unlab_dp_assms:
  assumes I: "dpp_spec I"
    and fin: "\<forall>p\<in>set (dp_assms a prf). finite_dpp_assm p"
    and ok: "isOK (check_dp_termination_proof_with_unlab a i I dpp prf)"
  shows "finite_dpp (dpp_ops.dpp I dpp)"
  by (rule check_dp_termination_proof_with_dp_assms [OF I _ ok [unfolded check_dp_termination_proof_with_unlab_def]], insert fin unlab_to_split_proc_assms, auto)

lemma check_dp_termination_proof_with_unlab:
  assumes "dpp_spec I"
    and "isOK (check_dp_termination_proof_with_unlab False i I dpp prf)"
  shows "finite_dpp (dpp_ops.dpp I dpp)"
  using check_dp_termination_proof_with_unlab_dp_assms[OF assms(1) _ assms(2)] by simp

end

