(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Critical_Pair_Closure
imports 
  "$AFP/Abstract-Rewriting/Abstract_Rewriting"
  Critical_Pairs
begin

(* Lemma 2.3 in Huet: Confluent Reductions *)
lemma partially_localize_CR: "(CR r) = (\<forall> x y z. (x, y) \<in> r \<and> (x, z) \<in> r^* \<longrightarrow> (y, z) \<in> join r)" 
proof
  assume "CR r"
  thus "\<forall> x y z. (x, y) \<in> r \<and> (x, z) \<in> r^* \<longrightarrow> (y, z) \<in> join r" by auto
next
  assume 1:"\<forall> x y z. (x, y) \<in> r \<and> (x, z) \<in> r^* \<longrightarrow> (y, z) \<in> join r" 
  show "CR r"
  proof
    fix a b c
    assume 2:"a \<in> UNIV" and 3:"(a, b) \<in> r^*" and 4:"(a, c) \<in> r^*"
    then obtain n  where "(a,c) \<in> r^^n" using rtrancl_is_UN_relpow by fast
    with 2 3 show "(b,c) \<in> join r" 
    proof (induct n arbitrary: a b c)
      case 0 thus ?case by auto
    next
      case (Suc m)
      from Suc(4) obtain d where ad:"(a,d) \<in> r^^m" and dc:"(d,c) \<in> r" by auto
      from Suc(1)[OF Suc(2) Suc(3) ad] have "(b,d) \<in> join r" .
      with 1 dc joinE joinI[of b _ r c] join_rtrancl_join show ?case by metis
    qed
  qed
qed

definition strongly_confluent_on :: "'a rel \<Rightarrow> 'a set \<Rightarrow> bool" where 
  "strongly_confluent_on r A \<equiv> \<forall>x \<in> A.  \<forall>y z. (x, y) \<in> r \<and> (x, z) \<in> r \<longrightarrow> (\<exists>u. (y,u) \<in> r^* \<and> (z,u) \<in> r^=)"

abbreviation strongly_confluent :: "'a rel \<Rightarrow> bool" where
  "strongly_confluent r \<equiv> strongly_confluent_on r UNIV"

(* Lemma 2.5 in Huet: Confluent Reductions *)
lemma strong_confluence_imp_CR: 
  assumes "strongly_confluent r"
  shows "CR r"
proof -
  from assms have 
    1:"\<And> x y z n. (x,y) \<in> r^= \<and> (x,z) \<in> r^^n \<Longrightarrow> \<exists> u. (y,u) \<in> r^* \<and> (z,u) \<in> r^=" 
    proof -
      fix x y z n
      assume as:"(x, y) \<in> r\<^sup>= \<and> (x, z) \<in> r ^^ n"
      show "\<exists> u. (y,u) \<in> r^* \<and> (z,u) \<in> r^="
      using as proof (induct n arbitrary: x y z)
        case 0
        thus ?case by auto
      next
        case (Suc m)
        from Suc(2) obtain w where xw:"(x,w) \<in> r^^m" and wz:"(w,z) \<in> r" by auto
        from xw Suc(2) have "(x,y) \<in> r^= \<and> (x,w) \<in> r^^m" by auto
        from Suc(1)[OF this] have "\<exists> u. (y,u) \<in> r^* \<and> (w,u) \<in> r^=" .
        then obtain u where u:"(y,u) \<in> r^* \<and> (w,u) \<in> r^=" ..
        with assms wz UNIV_I Un_iff pair_in_Id_conv rtrancl.rtrancl_into_rtrancl rtrancl_trans 
        show ?case unfolding strongly_confluent_on_def by metis
      qed
    qed
  have "(\<forall> x y z. (x, y) \<in> r \<and> (x, z) \<in> r^* \<longrightarrow> (y, z) \<in> join r)" 
  proof (intro allI, intro impI)
    fix x y z
    assume a:"(x, y) \<in> r \<and> (x, z) \<in> r^*"
    show "(y,z) \<in> join r" 
    proof (cases "x = y")
      case True
      with a show ?thesis by blast
    next
      case False
      with 1 a show ?thesis by blast
    qed
  qed
  thus "CR r" using partially_localize_CR by blast
qed

definition strongly_closed :: "('f,string) trs \<Rightarrow> bool" where 
  "strongly_closed R \<equiv> \<forall> (b,p,q) \<in> critical_pairs R R. \<exists> s r. 
                         (p,r) \<in> (rstep R)^* \<and> (q,r) \<in> (rstep R)^= 
                       \<and> (p,s) \<in> (rstep R)^= \<and> (q,s) \<in> (rstep R)^*"

(* TODO: move? *)
definition
  linear_trs :: "('f, 'v) trs \<Rightarrow> bool"
where
  "linear_trs R \<equiv> \<forall>(l, r)\<in>R. linear_term l \<and> linear_term r"

(* Lemma 3.2 in Huet: Confluent Reductions *)
lemma strongly_closed_linear_imp_strongly_confluent: 
  assumes sc:"strongly_closed R"
  and lin:"linear_trs R"
  shows "strongly_confluent (rstep R)"
unfolding strongly_confluent_on_def 
proof (simp only: Set.ball_UNIV, intro allI, intro impI)
  let ?R = "rstep R"
  fix x y z
  assume "(x, y) \<in> ?R \<and> (x, z) \<in> ?R"
  then obtain l1 r1 p1 \<sigma>1 l2 r2 p2 \<sigma>2 where 
    xy:"(x,y) \<in> rstep_r_p_s R (l1,r1) p1 \<sigma>1" and 
    xz:"(x,z) \<in> rstep_r_p_s R (l2,r2) p2 \<sigma>2" 
    using rstep_iff_rstep_r_p_s by metis
  from xy[unfolded rstep_r_p_s_def'] 
  have p1: "p1 \<in> poss x" and lr1: "(l1,r1) \<in> R" and \<sigma>1: "x |_ p1 = l1 \<cdot> \<sigma>1"
    and y: "y = replace_at x p1 (r1 \<cdot> \<sigma>1)" by auto
  from xz[unfolded rstep_r_p_s_def'] 
  have p2: "p2 \<in> poss x" and lr2: "(l2,r2) \<in> R" and \<sigma>2: "x |_ p2 = l2 \<cdot> \<sigma>2"
    and z: "z = replace_at x p2 (r2 \<cdot> \<sigma>2)" by auto
  show " \<exists>u. (y, u) \<in> ?R^* \<and> (z, u) \<in> ?R^=" 
  proof (cases "p1 \<bottom> p2")
    case True note p1p2 = this
    from y z have u:"replace_at y p2 (r2 \<cdot> \<sigma>2) = replace_at z p1 (r1 \<cdot> \<sigma>1)" 
      using parallel_replace_at[OF p1p2 p1 p2] by auto
    from parallel_poss_replace_at[OF p1p2 p1] y p2 have p2y:"p2 \<in> poss y" by auto
    from \<sigma>2 parallel_replace_at_subt_at[OF p1p2 p1 p2] y have 
      "y |_ p2 = l2 \<cdot> \<sigma>2" by auto
    with ctxt_supt_id[OF p2y] have "y = (ctxt_of_pos_term p2 y)\<langle>l2 \<cdot> \<sigma>2\<rangle>" by auto
    from rstepI[OF lr2 this] have 1:"(y,(ctxt_of_pos_term p2 y)\<langle>r2 \<cdot> \<sigma>2\<rangle>) \<in> ?R" by auto
    from p1p2 have p2p1:"p2 \<bottom> p1" using parallel_pos_sym by auto
    from parallel_poss_replace_at[OF p2p1 p2] z p1 have p1z:"p1 \<in> poss z" by auto
    from \<sigma>1 parallel_replace_at_subt_at[OF p2p1 p2 p1] z have 
      "z |_ p1 = l1 \<cdot> \<sigma>1" by auto
    with ctxt_supt_id[OF p1z] have "z = (ctxt_of_pos_term p1 z)\<langle>l1 \<cdot> \<sigma>1\<rangle>" by auto
    from rstepI[OF lr1 this] have "(z,(ctxt_of_pos_term p1 z)\<langle>r1 \<cdot> \<sigma>1\<rangle>) \<in> ?R" by auto
    with 1 u show ?thesis by auto
  next
    case False
    {
      fix p1 p2 l1 r1 \<sigma>1 y l2 r2 \<sigma>2 z
      assume 
        le:"(p1 :: pos) \<le> p2" and
        p1: "p1 \<in> poss x" and \<sigma>1: "x |_ p1 = l1 \<cdot> \<sigma>1" and y: "y = replace_at x p1 (r1 \<cdot> \<sigma>1)" and lr1: "(l1,r1) \<in> R" and
        p2: "p2 \<in> poss x" and \<sigma>2: "x |_ p2 = l2 \<cdot> \<sigma>2" and z: "z = replace_at x p2 (r2 \<cdot> \<sigma>2)" and lr2: "(l2,r2) \<in> R"
      then obtain v where v:"p2 = p1 <#> v" using less_eq_pos_def by auto
      hence l2v:"(l1 \<cdot> \<sigma>1) |_ v = l2 \<cdot> \<sigma>2" using \<sigma>1 \<sigma>2 p1 p2 by auto
      from v z \<sigma>1 have zp1:"z = replace_at x p1 (replace_at (l1 \<cdot> \<sigma>1) v (r2 \<cdot> \<sigma>2))" 
        using context_of_pos_term_append[OF p1] ctxt_compose_apply_distrib by force
      have 
        "(v \<in> poss l1 \<and> is_Fun (l1 |_ v) \<and> l1 |_ v \<cdot> \<sigma>1 = l2 \<cdot> \<sigma>2) \<or> 
        (\<exists> v1 v2 w. 
        v = v1 <#> v2 \<and> 
        v1 \<in> poss l1 \<and> l1 |_ v1 = Var w \<and> 
        \<sigma>1 w |_ v2 = l2 \<cdot> \<sigma>2)" 
      proof (cases "v \<in> poss l1 \<and> is_Fun (l1 |_ v)")
        case True
        hence "v \<in> poss l1" ..
        with l2v subt_at_subst[OF this] have "l1 |_ v \<cdot> \<sigma>1 = l2 \<cdot> \<sigma>2" by auto
        with True show ?thesis by fast
      next
        case False       
        then obtain v1 v2 w where 
          v12:"v = v1 <#> v2" and v1:"v1 \<in> poss l1" and l1v1:"l1 |_ v1 = Var w"
          using \<sigma>1 p2 pos_into_subst[of l1 \<sigma>1 "x |_ p1" v] poss_append_poss v by auto
        moreover hence "\<sigma>1 w |_ v2 = l2 \<cdot> \<sigma>2" 
          using l2v v1
            subt_at_append[OF poss_imp_subst_poss[OF v1, of \<sigma>1], of v2] by auto
        ultimately show ?thesis by blast
      qed
      hence "(\<exists>u. (y, u) \<in> (rstep R)^* \<and> (z, u) \<in> (rstep R)^=) \<and>
        (\<exists>u. (y, u) \<in> (rstep R)^= \<and> (z, u) \<in> (rstep R)^*)" 
      proof
      assume "\<exists> v1 v2 w. 
        v = v1 <#> v2 \<and> 
        v1 \<in> poss l1 \<and> l1 |_ v1 = Var w \<and> 
        \<sigma>1 w |_ v2 = l2 \<cdot> \<sigma>2"
        then obtain v1 v2 w where 
          v12:"v = v1 <#> v2" and 
          w:"l1 |_ v1 = Var w" and
          v1:"v1 \<in> poss l1" and
          wv2:"\<sigma>1 w |_ v2 = l2 \<cdot> \<sigma>2"
           by blast
        have v1l1\<sigma>1:"v1 \<in> poss (l1 \<cdot> \<sigma>1)" using v1 by simp
        let ?\<sigma>3 = "\<lambda> y. if y = w then replace_at (\<sigma>1 y) v2 (r2 \<cdot> \<sigma>2) else \<sigma>1 y"
        let ?cr = "replace_at x p1 (r1 \<cdot> ?\<sigma>3)"
        from lin lr1 have linl1:"linear_term l1" unfolding linear_trs_def by fast
        from linear_term_replace_in_subst[OF linl1 v1 w, of \<sigma>1 ?\<sigma>3, simplified]
        have "l1 \<cdot> ?\<sigma>3 = replace_at (l1 \<cdot> \<sigma>1) v1 (?\<sigma>3 w)" by auto
        also have "... = replace_at (l1 \<cdot> \<sigma>1) v1 (replace_at (\<sigma>1 w) v2 (r2 \<cdot> \<sigma>2))" by auto
        also have "... = replace_at (l1 \<cdot> \<sigma>1) v (r2 \<cdot> \<sigma>2)" 
          using v12 v1 w
            context_of_pos_term_append[OF v1l1\<sigma>1] 
            ctxt_compose_apply_distrib by auto
        finally have "l1 \<cdot> ?\<sigma>3 = replace_at (l1 \<cdot> \<sigma>1) v (r2 \<cdot> \<sigma>2)" .
        with zp1 have "z = replace_at x p1 (l1 \<cdot> ?\<sigma>3)" by auto 
        hence zcr:"(z, ?cr) \<in> ?R" using subset_rstep lr1 by auto
        have ycr:"(y, ?cr) \<in> ?R^=" 
        proof (cases "w \<in> vars_term r1")
          case False
          hence "r1 \<cdot> ?\<sigma>3 = r1 \<cdot> \<sigma>1" by (induct r1, auto)
          hence "?cr = y" using y by auto
          thus ?thesis by auto
        next
          case True
          obtain q where q:"q \<in> poss r1" and qw:"r1 |_ q = Var w" 
            using  supteq_imp_subt_at[OF supteq_Var[OF True]] by auto
          from lin lr1 have linr1:"linear_term r1" unfolding linear_trs_def by fast
          have qr1\<sigma>1:"q \<in> poss (r1 \<cdot> \<sigma>1)" using q by simp
          from linear_term_replace_in_subst[OF linr1 q qw, of \<sigma>1 ?\<sigma>3, simplified]
          have "r1 \<cdot> ?\<sigma>3 = replace_at (r1 \<cdot> \<sigma>1) q (?\<sigma>3 w)" by auto
          also have "... = replace_at (r1 \<cdot> \<sigma>1) q (replace_at (\<sigma>1 w) v2 (r2 \<cdot> \<sigma>2))" by auto
          also have "... = replace_at (r1 \<cdot> \<sigma>1) (q <#> v2) (r2 \<cdot> \<sigma>2)" 
            using q qw
              context_of_pos_term_append[OF qr1\<sigma>1]
              ctxt_compose_apply_distrib by auto
          finally have s3:"r1 \<cdot> ?\<sigma>3 = replace_at (r1 \<cdot> \<sigma>1) (q <#> v2) (r2 \<cdot> \<sigma>2)" .
          from wv2 qw have "(r1 \<cdot> \<sigma>1) |_ (q <#> v2) = l2 \<cdot> \<sigma>2" 
            using q subt_at_append[OF qr1\<sigma>1, of v2] by auto
          hence "r1 \<cdot> \<sigma>1 = (ctxt_of_pos_term (q <#> v2) (r1 \<cdot> \<sigma>1))\<langle>l2 \<cdot> \<sigma>2\<rangle>"
            using qr1\<sigma>1 ctxt_supt_id \<sigma>1 p2 poss_append_poss q qw subt_at_subst v v1 v12 w by metis
          from rstepI[OF lr2 this] s3 have "(r1 \<cdot> \<sigma>1, r1 \<cdot> ?\<sigma>3) \<in> ?R" by simp
          thus ?thesis using y by blast
        qed
        from zcr ycr show ?thesis by blast
      next
        assume "v \<in> poss l1 \<and> is_Fun (l1 |_ v) \<and> l1 |_ v \<cdot> \<sigma>1 = l2 \<cdot> \<sigma>2"
        hence 
          vl1:"v \<in> poss l1" and 
          l1v:"is_Fun (l1 |_ v)" and 
          l1vs1: "l1 |_ v \<cdot> \<sigma>1 = l2 \<cdot> \<sigma>2" by auto
        from mgu_var_disjoint_string_complete[OF l1vs1]
        obtain \<mu>1 \<mu>2 \<delta> where mgu: "mgu_var_disjoint_string (l1 |_ v) l2 = Some (\<mu>1,\<mu>2)" 
          and d1:"\<sigma>1 = \<mu>1 \<circ>s \<delta>" 
          and d2:"\<sigma>2 = \<mu>2 \<circ>s \<delta>" 
          and mu12:"l1 |_ v \<cdot> \<mu>1 = l2 \<cdot> \<mu>2"
          by fast
        let ?s = "r1 \<cdot> \<mu>1"
        let ?t = "(ctxt_of_pos_term v l1 \<cdot>c \<mu>1)\<langle>r2 \<cdot> \<mu>2\<rangle>"
        from critical_pairsI[OF lr1 lr2 _ l1v mgu, of "ctxt_of_pos_term v l1" ?s ?t
          "ctxt_of_pos_term v l1 = \<box>"] ctxt_supt_id[OF vl1] have
          "(ctxt_of_pos_term v l1 = \<box>,?s,?t) \<in> critical_pairs R R" by auto
        with sc[unfolded strongly_closed_def] obtain m n where
          m: "(?s,m) \<in> ?R^* \<and> (?t,m) \<in> ?R^=" and n: "(?s,n) \<in> ?R^= \<and> (?t,n) \<in> ?R^*" by blast
        from d2 d1 have d2:"?t \<cdot> \<delta> = replace_at (l1 \<cdot> \<sigma>1) v (r2 \<cdot> \<sigma>2)" 
          using ctxt_compose_subst_compose_distrib ctxt_of_pos_term_subst[OF vl1] 
            subst_apply_term_ctxt_apply_distrib subst_compose_apply_term_distrib by metis
        from d1 have d1:"?s \<cdot> \<delta> = r1 \<cdot> \<sigma>1" by auto
        from d1 y have ys:"y = replace_at x p1 (?s \<cdot> \<delta>)" by auto
        from d2 z zp1 have zt:"z = replace_at x p1 (?t \<cdot> \<delta>)" by auto
        from m have minst:"(?s \<cdot> \<delta>, m \<cdot> \<delta>) \<in> ?R^* \<and> (?t \<cdot> \<delta>, m \<cdot> \<delta>) \<in> ?R^="
          using rstep.subst rstep_id rstep_union rsteps_closed_subst by metis
        from n have ninst:"(?s \<cdot> \<delta>, n \<cdot> \<delta>) \<in> ?R^= \<and> (?t \<cdot> \<delta>, n \<cdot> \<delta>) \<in> ?R^*" 
          using rstep.subst rstep_id rstep_union rsteps_closed_subst by metis
        let ?cr1 = "replace_at x p1 (m \<cdot> \<delta>)"
        from minst ys have ycr1:"(y,?cr1) \<in> ?R^*" using rsteps_closed_ctxt by auto
        from minst zt have zcr1:"(z,?cr1) \<in> ?R^=" by auto
        let ?cr2 = "replace_at x p1 (n \<cdot> \<delta>)"
        from ninst ys have ycr2:"(y,?cr2) \<in> ?R^=" by auto
        from ninst zt have zcr2:"(z,?cr2) \<in> ?R^*" using rsteps_closed_ctxt by auto
        from ycr1 zcr1 ycr2 zcr2 show ?thesis by blast
      qed
    } note wlog = this
    from False have "(p1 \<le> p2 \<or> p2 \<le> p1)" using parallel_pos by simp
    thus ?thesis proof 
      assume le:"p1 \<le> p2"
      from wlog[OF le p1 \<sigma>1 y lr1 p2 \<sigma>2 z lr2] show ?thesis by fast
    next
      assume le:"p2 \<le> p1"
      from wlog[OF le p2 \<sigma>2 z lr2 p1 \<sigma>1 y lr1] show ?thesis by blast
    qed
  qed
qed

corollary strongly_closed_linear_CR:
  assumes "strongly_closed R"
    and "linear_trs R"
  shows "CR (rstep R)"
  using assms strongly_closed_linear_imp_strongly_confluent strong_confluence_imp_CR by auto

end

