(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Rewriting_Impl
imports 
  Innermost_Usable_Rules_Impl
  Innermost_Impl
  Critical_Pairs_Impl
  Rewriting
begin

definition check_rewrite_common_preconditions where
  "check_rewrite_common_preconditions I U_opt st ss t' lr p sound dpp \<equiv> do {
     let R = dpp_ops.rules I dpp;
     let s = fst st;
     let t = snd st;
     let tp = t |_ p;
     let U = (case U_opt of Some U \<Rightarrow> U | None => inn_usable_rules_pair I dpp (s,tp));
     check_subseteq U R <+? (\<lambda> lr. shows_rule lr +@+ shows '' is not a rule of the rewrite system '');
     let urc = is_ur_closed_impl_dpp_mv I dpp U;
     let check_urc = (\<lambda> S t. check (urc S t) (shows ''term '' +@+ shows t +@+ shows '' is not closed under usable rules''));
     let nfs = dpp_ops.nfs I dpp;
     check_allm (\<lambda> (l,r). check (is_Fun l) (shows ''lhss must not be variables'')) U;
     check (wf_rule lr) (shows_rule lr +@+ shows '' is not a well formed rule'');
     (if nfs \<and> sound then succeed else (check_subseteq (vars_term_list tp) (vars_term_list s) <+? (\<lambda> x. shows ''variable condition in pair violated'')));
     check_urc ss tp;
     check_allm (\<lambda>(l,r). check_urc (args l) r) U;
     check_critical_pairs_innermost U <+? (\<lambda> s. shows ''problem in showing UNF of usable rules'' +@+ shows_nl +@+ s);
     check_allm (\<lambda>(b,s,t). check (s = t) (shows ''non-trivial critical pair between rule to rewrite and usable rules'')) (critical_pairs_impl [lr] U)
   }"

lemma check_rewrite_common_preconditions: assumes I: "dpp_spec I"
  and ok: "isOK(check_rewrite_common_preconditions I U_opt (s,t) ss t' lr p sound dpp)"
  and inn: "NF_terms (set (dpp_ops.Q I dpp)) \<subseteq> NF_trs (set (dpp_ops.rules I dpp))"
  shows "R_Q_U_ecap (set (dpp_ops.rules I dpp)) (set (dpp_ops.Q I dpp)) icap'"
    "(\<exists> U. R_Q_U_ecap.rewrite_common_preconditions (set (dpp_ops.rules I dpp)) U (set (dpp_ops.Q I dpp)) icap' 
    s ss t t' lr p (dpp_ops.nfs I dpp) sound)"
proof -
  interpret dpp_spec I by fact
  let ?R = "set (dpp_ops.rules I dpp)"
  let ?Q = "set (dpp_ops.Q I dpp)"
  let ?U' = "case U_opt of None \<Rightarrow> inn_usable_rules_pair I dpp (s,t |_ p) | Some U \<Rightarrow> U"
  def U \<equiv> ?U'
  let ?U = "set U"
  let ?nfs = "dpp_ops.nfs I dpp"
  note ok = ok[unfolded check_rewrite_common_preconditions_def Let_def is_ur_closed_impl_dpp_mv[OF I] fst_conv snd_conv,
    folded U_def, simplified] 
  show "R_Q_U_ecap ?R ?Q icap'" 
    by (unfold_locales, rule icap, rule inn)
  interpret R_Q_U_ecap ?R ?U ?Q icap' by fact
  from ok have subset: "?U \<subseteq> ?R" by auto
  show "\<exists> U. R_Q_U_ecap.rewrite_common_preconditions ?R U ?Q icap' s ss t t' lr p ?nfs sound"
  proof (rule exI[of _ ?U], unfold rewrite_common_preconditions_def, intro conjI allI impI)
    from ok have ok': "isOK(check_critical_pairs_innermost U)" by simp
    from NF_Q_R NF_anti_mono[OF rstep_mono[OF subset]] have NFQU: "NFQ \<subseteq> NF_trs (set U)" by auto
    {
      fix l r
      assume "(l,r) \<in> ?U"
      with ok have "is_Fun l" by force
    }
    from check_critical_pairs_innermost[OF ok' NFQU this]
    have "CR (qrstep ?nfs ?Q ?U)" .
    from CR_imp_UNF[OF this] show "UNF (qrstep ?nfs ?Q ?U)" .
  qed (insert ok, auto)
qed
      

(* TODO: one might automatically put st' into strict pairs,
   if st is not strict and there are strict rules *)
definition
  rewriting_proc ::
    "('dpp, 'f :: {key,show}, string) dpp_ops \<Rightarrow> 
     ('f,string)rules option \<Rightarrow> ('f,string)rule \<Rightarrow> ('f,string)rule \<Rightarrow> ('f,string)rule \<Rightarrow> ('f,string)rule \<Rightarrow> pos \<Rightarrow> 'dpp proc"
where
  "rewriting_proc I U st st' st'' lr p dpp \<equiv> 
    check_return (do {
     let s = fst st;
     let t' = snd st';
     check_rstep' (dpp_ops.rules I dpp) p lr (snd st) t';
     check (dpp_ops.NFQ_subset_NF_rules I dpp) (shows ''innermost rewriting required'');
     check_rewrite_common_preconditions I U st [s] t' lr p True dpp;
     check (st' =v st'') (shows ''the rule '' +@+ shows_rule st' +@+ shows '' is not a renamed variant of '' +@+ shows_rule st'');
     check (s = fst st') (shows ''left-hand sides of old and new pair differ'');
     check (st \<in> set (dpp_ops.P I dpp) \<or> dpp_ops.R I dpp = []) (shows ''strict DP or no strict rules required'');
     check (dpp_ops.nfs I dpp \<or> dpp_ops.wwf_rules I dpp) (shows ''well-formed rules or normal subst. required'')
   })
   (dpp_ops.replace_pair I dpp st [st''])"    

lemma variable_renamed_variant: assumes eqv: "st1 =v st2" and st: "st \<in> replace st0 ({st1}) P"
  shows "\<exists> st'. st' \<in> replace st0 ({st2}) P \<and> st =v st'"
proof -
  note rep_def = Util.replace_def
  have str: "st =v st" by simp
  show ?thesis
  proof (cases "st0 \<in> P")
    case False
    with st str show ?thesis unfolding rep_def 
      by (intro exI[of _ st], auto)
  next
    case True 
    with st[unfolded rep_def]
    have "st \<in> P - {st0} \<union> {st1}" by simp
    thus ?thesis
    proof
      assume "st \<in> P - {st0}" 
      with True str show ?thesis unfolding rep_def
        by (intro exI[of _ st], auto)
    next
      assume "st \<in> {st1}"
      with True eqv show ?thesis unfolding rep_def
        by (intro exI[of _ "st2"], auto)
    qed
  qed
qed
 
lemma rewriting_proc: assumes I: "dpp_spec I"
 shows "dpp_spec.sound_proc_impl I (rewriting_proc I (U_opt :: ('f :: {key,show},string)rules option)
  st st' st'' lr p)"
proof -
  from assms interpret dpp_spec I .
  show ?thesis
  proof
    fix d d''    
    assume ok: "rewriting_proc I U_opt st st' st'' lr p d = Inr d''" and fin: "finite_dpp (dpp d'')"
    let ?P = "set (P d)"
    let ?Pw = "set (Pw d)"
    let ?Q = "set (Q d)"
    let ?R = "set (R d)"
    let ?Rw = "set (Rw d)"
    let ?Rb = "?R \<union> ?Rw"
    let ?Rb' = "set (rules d)"
    let ?nfs = "NFS d"
    let ?m = "M d"
    obtain s t where st: "st = (s,t)" by force
    obtain l r where lr: "lr = (l,r)" by force
    obtain s' t' where st': "st' = (s',t')" by force
    obtain s'' t'' where st'': "st'' = (s'',t'')" by force
    note ok = ok[unfolded rewriting_proc_def st lr st' st'' Let_def dpp_spec_sound, simplified]
    from ok have inn: "NF_terms ?Q \<subseteq> NF_trs ?Rb'" by auto
    from ok obtain \<sigma> where rewrite: "(t, t') \<in> rstep_r_p_s ?Rb' (l, r) p \<sigma>" by auto
    from ok have common: "isOK (check_rewrite_common_preconditions I U_opt (s, t) [s] t' (l, r) p True d)" by auto
    note common = check_rewrite_common_preconditions[OF I this inn]
    obtain d' where d': "d' = replace_pair d (s,t) [(s,t')]" by auto
    from ok have d'': "d'' = replace_pair d (s,t) [(s'',t'')]" by auto
    from ok have eqv: "(s,t') =v (s'',t'')" by simp
    note variant = variable_renamed_variant[OF this, of _ "(s,t)"]
    from fin have fin: "finite_dpp (dpp d')" unfolding d' d'' replace_pair_sound
      by (rule finite_dpp_rename_vars, insert variant[of _ ?P] variant[of _ ?Pw], auto)
    have d': "dpp d' = (?nfs,?m,replace (s,t) {(s,t')} ?P, replace (s,t) {(s,t')} ?Pw, ?Q, ?R, ?Rw)" 
      unfolding d' unfolding replace_pair_sound by simp
    note fin = fin[unfolded this]
    from common(2) obtain U where comm: "R_Q_U_ecap.rewrite_common_preconditions ?Rb' U (set (Q d)) icap' s [s] t t' (l, r) p (NFS d) True"
      by blast
    interpret R_Q_U_ecap ?Rb' U ?Q icap' by (rule common(1))
    show "finite_dpp (dpp d)" unfolding dpp_spec_sound
      by (rule rewriting_proc[OF fin comm rewrite], insert ok, auto)
  qed
qed

definition check_nfc :: "bool \<Rightarrow> ('f :: show, 'v :: show)rules \<Rightarrow> ('f,'v)term list \<Rightarrow> ('f,'v)term list \<Rightarrow> bool \<Rightarrow> ('f,'v)term \<Rightarrow> shows check" where 
  "check_nfc inn R Q ss nfs t \<equiv> check inn (shows ''we do not have exactly innermost and currently no better nfc-condition test is implemented'')"
  

lemma check_nfc: assumes inn: "NF_terms (set Q) \<subseteq> NF_trs (set R)"
  and ok: "isOK(check_nfc (NF_trs (set R) \<subseteq> NF_terms (set Q)) R Q ss nfs t)"
  and wf: "\<And> l r. nfs \<Longrightarrow> (l,r) \<in> set R \<Longrightarrow> is_Fun l"
  and sub: "t \<rhd> u"
  shows "nfc (set R) (set Q) (set ss) u nfs"
  by (rule innermost_imp_nfc[OF _ wf], insert ok[unfolded check_nfc_def] inn, auto)

definition
  rewriting_complete_proc ::
    "('dpp, 'f :: {key,show}, string) dpp_ops \<Rightarrow> 
     ('f,string)rules option \<Rightarrow> ('f,string)rule \<Rightarrow> ('f,string)rule \<Rightarrow> ('f,string)rule \<Rightarrow> ('f,string)rule \<Rightarrow> pos \<Rightarrow> 'dpp proc"
where
  "rewriting_complete_proc I U_opt st st' st'' lr p dpp \<equiv> 
    check_return (do {
     let s = fst st;
     let t = snd st;
     let t' = snd st';
     let R = dpp_ops.rules I dpp;
     check_rstep' R p lr t t';
     check (dpp_ops.NFQ_subset_NF_rules I dpp) (shows ''innermost rewriting required'');     
     check_rewrite_common_preconditions I U_opt st (args s) t' lr p False dpp;
     check (is_Fun s) (shows ''lhs of pair must not be variable'');
     check (st' =v st'') (shows ''the rule '' +@+ shows_rule st' +@+ shows '' is not a renamed variant of '' +@+ shows_rule st'');
     check (s = fst st') (shows ''left-hand sides of old and new pair differ'');
     let Q = dpp_ops.Q I dpp;
     let inn = isOK(check_NF_terms_subset (map fst R) Q);
     check_nfc inn R Q (args s) (dpp_ops.nfs I dpp) (t |_ p);
     check_allm (\<lambda> (l,r). check (is_Fun l) (shows ''lhss must not be variables'')) R;
     (if is_Fun t then check (\<not> dpp_spec.is_defined I dpp (the (root t))) (shows ''root of '' +@+ shows t +@+ shows '' must not be defined'')
       else succeed)
   })
   (dpp_ops.replace_pair I dpp st [st''])"  
 
lemma rewriting_complete_proc: assumes I: "dpp_spec I"
  and ok: "rewriting_complete_proc I U_opt st st' st'' lr p d = return d''"
  and infin: "infinite_dpp (dpp_ops.nfs I d'', set (dpp_ops.pairs I d''), set (dpp_ops.Q I d''), set (dpp_ops.rules I d''))" 
    (is "infinite_dpp ?dpp''")
  shows "infinite_dpp (dpp_ops.nfs I d, set (dpp_ops.pairs I d), set (dpp_ops.Q I d), set (dpp_ops.rules I d))"
    (is "infinite_dpp ?dpp")
proof -
  interpret dpp_spec I by fact
  let ?R = "set (rules d)"
  let ?nfs = "NFS d"
  let ?Q = "set (Q d)"
  let ?P = "set (pairs d)"
  obtain s t where st: "st = (s,t)" by force
  obtain l r where lr: "lr = (l,r)" by force
  obtain s' t' where st': "st' = (s',t')" by force
  obtain s'' t'' where st'': "st'' = (s'',t'')" by force
  note ok = ok[unfolded rewriting_complete_proc_def st lr st' st'' Let_def, simplified]
  obtain d' where d': "d' = replace_pair d (s,t) [(s,t')]" by auto  
  let ?dpp' = "(dpp_ops.nfs I d', set (dpp_ops.pairs I d'), set (dpp_ops.Q I d'), set (dpp_ops.rules I d'))"
  from ok have d'': "d'' = replace_pair d (s,t) [(s'',t'')]" by auto
  from ok have inn: "NF_terms ?Q \<subseteq> NF_trs ?R" by auto
  from ok obtain \<sigma> where rewrite: "(t, t') \<in> rstep_r_p_s ?R (l, r) p \<sigma>" by auto
  from ok have "isOK (check_rewrite_common_preconditions I U_opt (s, t) (args s) t' (l, r) p False d)" by auto
  note common = check_rewrite_common_preconditions[OF I this inn]
  note id = replace_pair_sound[unfolded dpp_sound]
  note id1 = id[of d "(s,t)" "[(s,t')]"]
  have id1: "?dpp' = (?nfs, replace (s,t) {(s,t')} ?P, ?Q, ?R)" unfolding d' using id1 
    by (auto simp: replace_def)
  from ok have "(s,t') =v (s'',t'')" by simp
  hence eqv: "(s'',t'') =v (s,t')" unfolding eq_rule_mod_vars_def by simp
  note variant = variable_renamed_variant[OF this]
  note id2 = id[of d "(s,t)" "[(s'',t'')]"]
  have id2: "?dpp'' = (?nfs, replace (s,t) {(s'',t'')} ?P, ?Q, ?R)" unfolding d'' using id2
    by (auto simp: replace_def)
  from infin have infin: "infinite_dpp ?dpp'" unfolding id1 id2
    by (rule infinite_dpp_rename_vars, insert variant, auto)
  from common(2) obtain U where comm: "R_Q_U_ecap.rewrite_common_preconditions ?R U (set (Q d)) icap' s (args s) t t' (l, r) p (NFS d) False"
    by blast
  interpret R_Q_U_ecap ?R U ?Q icap' by (rule common(1))
  have nfc: "\<And> u. t |_ p \<rhd> u \<Longrightarrow> nfc ?R ?Q (set (args s)) u ?nfs"
    by (rule check_nfc[OF inn], insert ok, auto)
  show "infinite_dpp ?dpp"
    by (rule rewriting_proc_complete[OF infin[unfolded id1] comm rewrite _ nfc], insert ok, auto)
qed

end
