(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Size_Change_Termination_Processors
imports
  Size_Change_Termination
  Edg
  Generic_Usable_Rules
  Subterm_Criterion
  Dependency_Graph
begin

(* get_arg delivers the whole term (for 0), or otherwise the i-th argument, where 
   arguments are here indexed starting by 1 *)
fun get_arg :: "('f, 'v) term \<Rightarrow> nat \<Rightarrow> ('f, 'v) term" where
  "get_arg t 0 = t"
| "get_arg t (Suc n) = args t ! n"

lemma (in redtriple) sct_semantics: "sct_semantics S (NS \<union> NST)"
  by (rule sct_semantics.intro, rule SN, rule both.compat)

lemma generic_sct_redtriple:
   fixes P :: "('f,'v)trs" and R :: "('f,'v)rules" and Gs :: "(('f,'v)term \<times> ('f,'v)term \<times> 'a, nat) scg list"
  defines sts: "sts \<equiv> {(s,get_arg t j) | s t j . \<exists> i st ns. Scg (s,t,i) (s,t,i) st ns \<in> set Gs \<and> j \<in> snd ` set st \<union> snd ` set ns}"
  assumes tuple: "\<forall> (s,t) \<in> P. is_Fun s \<and> is_Fun t \<and> (the (root t)) \<notin> defs (set R)"
  and var_R: "\<And> l r. (l,r) \<in> set R \<Longrightarrow> is_Fun l"
  and checker: "usable_rules_checker checker"
  and U: "set U \<subseteq> NS"
  and redtriple: "ce_af_redtriple S NS NST \<pi>"
  and Ucheck: "checker nfs m (wwf_qtrs Q (set R)) \<pi> Q R U_opt sts = Some U"
  and graphs: "\<forall> (s,t) \<in> P. \<exists> stri non_stri. (
         Scg (s,t,info s t) (s,t,info s t) stri non_stri \<in> set Gs \<and>
         (\<forall> (i,j) \<in> set stri. i \<le> length (args s) \<and> j \<le> length (args t) \<and> (get_arg s i, get_arg t j) \<in> S) \<and> 
         (\<forall> (i,j) \<in> set non_stri. i \<le> length (args s) \<and> j \<le> length (args t) \<and> (get_arg s i, get_arg t j) \<in> NST))"
  and edg: "\<And> s t u v. ((s,t),(u,v)) \<in> DG nfs m P Q (set R) \<Longrightarrow> edg (s,t,info s t) (u,v,info u v)"  
  and check: "check_SCT (\<lambda> (s,t,i_st) (u,v,i_uv). edg (s,t,i_st) (u,v,i_uv)) Gs" (is "check_SCT ?conn Gs")
  and P: "P = Ps \<union> Pw" and R: "set R = set Rs \<union> set Rw"
  shows "finite_dpp (nfs,m,Ps,Pw,Q,set Rs,set Rw)"
proof -
  interpret ce_af_redtriple S NS NST \<pi>  by fact
  show ?thesis
  proof (rule finite_dpp_mono)
    show "finite_dpp (nfs,m,P,{},Q,set R,{})" unfolding finite_dpp_def
    proof (clarify)
      fix s t \<sigma>
      assume chain: "min_ichain (nfs,m,P, {}, Q, set R, {}) s t \<sigma>"
      have ce: "ce_af_redpair S NS \<pi>" ..
      let ?R = "qrstep nfs Q (set R)"
      let ?Rs = "?R^*"
      note checker[unfolded usable_rules_checker_def, rule_format, OF Ucheck ce U]
      then obtain I where I: "\<And> s t u \<sigma> \<tau>. (s,t) \<in> sts \<Longrightarrow> s \<cdot> \<sigma> \<in> NF_terms Q \<Longrightarrow> NF_subst nfs (s,t) \<sigma> Q \<Longrightarrow> (t \<cdot> \<sigma>, u \<cdot> \<tau>) \<in> ?Rs \<Longrightarrow> (m \<Longrightarrow> SN_on ?R {t \<cdot> \<sigma>}) \<Longrightarrow> (t \<cdot> I \<sigma>, u \<cdot> I \<tau>) \<in> NS^*" by blast
      let ?entry = "\<lambda> k. (s k,t k,info (s k) (t k))"
      let ?Is = "\<lambda> i. I (\<sigma> i)"
      let ?pairs = "\<lambda> k. (?entry k, (\<lambda> j. get_arg (s k) j \<cdot> ?Is k))"
      let ?R = "{(?pairs k, ?pairs (Suc k)) | k. True}"
      let ?S = "(NS \<union> NST \<union> S)\<^sup>* O S O (NS \<union> NST \<union> S)\<^sup>*"
      let ?NS = "(NS \<union> NST \<union> S)\<^sup>*"
      {
        fix k
        let ?sk = "\<lambda> i. get_arg (s k) i \<cdot> \<sigma> k"
        let ?tk = "\<lambda> i. get_arg (t k) i \<cdot> \<sigma> k"
        let ?sk1 = "\<lambda> i. args (s (Suc k)) ! i \<cdot> \<sigma> (Suc k)"
        let ?Isk = "\<lambda> i. get_arg (s k) i \<cdot> ?Is k"
        let ?Itk = "\<lambda> i. get_arg (t k) i \<cdot> ?Is k"
        let ?Isk1 = "\<lambda> i. get_arg (s (Suc k)) i \<cdot> ?Is (Suc k)"
        from chain have inP': "(s (Suc k), t (Suc k)) \<in> P" by (auto simp: ichain.simps)
        from chain have inP: "(s k, t k) \<in> P" by (auto simp: ichain.simps)
        with tuple have validDP: "is_Fun (t k) \<and> (the (root (t k))) \<notin> defs (set R)" by auto
        from chain inP tuple have validDP1: "is_Fun (s (Suc k))" by (auto simp: ichain.simps)
        from chain have steps: "(t k \<cdot> \<sigma> k, s (Suc k) \<cdot> \<sigma> (Suc k)) \<in> ?Rs" by (auto simp: ichain.simps)
        from chain have NF: "(s k \<cdot> \<sigma> k) \<in> NF_terms Q" by (auto simp: ichain.simps)
        from chain have NFs: "(s (Suc k) \<cdot> \<sigma> (Suc k)) \<in> NF_terms Q" by (auto simp: ichain.simps)
        from chain have nfs: "NF_subst nfs (s k, t k) (\<sigma> k) Q" by (auto simp: ichain.simps)
        from chain have nfs': "NF_subst nfs (s (Suc k), t (Suc k)) (\<sigma> (Suc k)) Q" by (auto simp: ichain.simps)
        from chain have tSN: "m \<Longrightarrow> SN_on (qrstep nfs Q (set R)) {t k \<cdot> \<sigma> k}" by (simp add: minimal_cond_def)
        have argSteps: "\<And> j. j \<le> length (args (t k)) \<Longrightarrow> (s k, get_arg (t k) j) \<in> sts \<Longrightarrow> (?Itk j, ?Isk1 j) \<in> NS^*"
        proof -
          fix j1
          assume j1: "j1 \<le> length (args (t k))" and mem: "(s k, get_arg (t k) j1) \<in> sts"
          note I = I[OF mem NF]
          show "(?Itk j1, ?Isk1 j1) \<in> NS^*"
          proof (cases j1)
            case 0
            show ?thesis
              by (rule I, unfold 0, insert steps tSN nfs, auto)
          next
            case (Suc j)
            with j1 have j: "j < length (args (t k))" by auto
            note I = I[unfolded Suc get_arg.simps]
            let ?tj = "args (t k) ! j"
            let ?sj = "args (s (Suc k)) ! j"
            from validDP obtain f ts where tk: "t k = Fun f ts \<and> (f,length ts) \<notin> defs (set R)" by (cases "t k", auto)
            from validDP1 obtain g ss where sk1: "s (Suc k) = Fun g ss" by (cases "s (Suc k)", auto) 
            let ?tss = "(map (\<lambda> t. t \<cdot> \<sigma> k) ts)"
            from tk have f: "(f, length ?tss) \<notin> defs (set R)" by simp
            let ?sss = "(map (\<lambda> s. s \<cdot> \<sigma> (Suc k)) ss)"
            from tSN have SN: "m \<Longrightarrow> SN_on (qrstep nfs Q (set R)) {Fun f ?tss}" using tk by auto
            from tk j have "?tj \<in> set ts" by auto
            hence "vars_term ?tj \<subseteq> vars_term (t k)" using tk by auto
            with nfs have nfs: "NF_subst nfs (s k, ?tj) (\<sigma> k) Q" unfolding NF_subst_def vars_rule_def by auto
            note I = I[OF this]
            from tk j have mem: "?tj \<cdot> \<sigma> k \<in> set ?tss" by auto
            from steps have "(Fun f ts \<cdot> \<sigma> k, Fun g ss \<cdot> \<sigma> (Suc k)) \<in> ?Rs" using tk sk1 by auto
            hence steps2: "(Fun f ?tss, Fun g ?sss) \<in> ?Rs" by auto
            have "\<exists> us. length us = length ?tss \<and> Fun g ?sss = Fun f us \<and> (\<forall> i < length ?tss. (?tss ! i, us ! i) \<in> ?Rs)" 
              by (rule nondef_root_imp_arg_qrsteps[OF steps2], insert var_R f[unfolded defs_def], force+)
            from this obtain us where nearly: "length us = length ?tss \<and> Fun g ?sss = Fun f us \<and> (\<forall> i < length ?tss. (?tss ! i, us ! i) \<in> ?Rs)" ..
            from nearly j tk sk1 have "(?tj \<cdot> \<sigma> k, ?sj \<cdot> \<sigma> (Suc k)) \<in> ?Rs" (is ?part1) by auto
            have SN2: "m \<Longrightarrow> SN_on (qrstep nfs Q (set R)) {?tj \<cdot> \<sigma> k}"
              by (rule SN_imp_SN_arg_gen[OF ctxt_closed_qrstep SN mem])
            note I = I[OF `?part1` SN2]
            thus ?thesis unfolding Suc by simp
          qed
        qed
        from steps have rsteps: "(t k \<cdot> \<sigma> k, s (Suc k) \<cdot> \<sigma> (Suc k)) \<in> (rstep (set R))^*"
          using rtrancl_mono[OF qrstep_subset_rstep] by auto
        have "((s k, t k), (s (Suc k), t (Suc k))) \<in> DG nfs m P Q (set R)"
          by (rule DG_I[OF inP inP' steps NF NFs nfs nfs' tSN])
        from edg[OF this]
        have conn: "?conn (?entry k) (?entry (Suc k))" by auto
        from inP obtain stri non_stri where graph: "Scg (?entry k) (?entry k) stri non_stri \<in> set Gs \<and> 
          (\<forall> (i,j) \<in> set stri. i \<le> length (args (s k)) \<and> j \<le> length (args (t k)) \<and> (get_arg (s k) i, get_arg (t k) j) \<in> S) \<and> 
          (\<forall> (i,j) \<in> set non_stri. i \<le> length (args (s k)) \<and> j \<le> length (args (t k)) \<and> (get_arg (s k) i, get_arg (t k) j) \<in> NST)" using graphs by force
        let ?g = "Scg (?entry k) (?entry k) stri non_stri"
        from graph have gGs: "?g \<in> set Gs" ..
        let ?check = "\<lambda> g. ((?entry k, ?Isk), (?entry (Suc k)), ?Isk1)
          \<in> sct_semantics.steps S (NS \<union> NST) ?conn g"
        have "?check ?g"
        proof (simp add: sct_semantics.steps.simps sct_semantics conn[simplified], intro conjI)
          {
            fix i j
            assume ij: "(i,j) \<in> set non_stri"
            with graph have i: "i \<le> length (args (s k))" and j: "j \<le> length (args (t k))"
              and NS: "(get_arg (s k) i, get_arg (t k) j) \<in> NST" by auto
            have "(s k, get_arg (t k) j) \<in> sts" unfolding sts using graph[THEN conjunct1] ij by force
            from argSteps[OF j this]
            have astepsNS: "(?Itk j, ?Isk1 j) \<in> NS^*"  .
            from NS subst_NST have non_strict: "(?Isk i, ?Itk j) \<in> NST" unfolding subst_closed_def using subst_closureI by blast
            have NS: "(?Isk i, ?Isk1 j) \<in> NST O NS^*" using astepsNS non_strict by blast
            have "(?Isk i, ?Isk1 j) \<in> ?NS" by (rule set_mp[OF _ NS], regexp)
          }
          thus "\<forall>ij\<in>set non_stri. (\<lambda>(i, j). (?Isk i, ?Isk1 j) \<in> ?NS) ij" by blast
        next
          {
            fix i j
            assume ij: "(i,j) \<in> set stri"
            with graph have i: "i \<le> length (args (s k))" and j: "j \<le> length (args (t k))"
              and S: "(get_arg (s k) i, get_arg (t k) j) \<in> S" by auto 
            have "(s k, get_arg (t k) j) \<in> sts" unfolding sts using graph[THEN conjunct1] ij by force
            from argSteps[OF j this]
            have astepsNS: "(?Itk j, ?Isk1 j) \<in> NS^*"  .
            from S subst_S have strict: "(?Isk i, ?Itk j) \<in> S" unfolding subst_closed_def using subst_closureI by blast
            have S: "(?Isk i, ?Isk1 j) \<in> S O NS^*" using strict astepsNS by blast
            have "(?Isk i, ?Isk1 j) \<in> ?S" by (rule set_mp[OF _ S], regexp)
          }
          thus "\<forall>ij\<in>set stri.(\<lambda>(i, j). (?Isk i, ?Isk1 j) \<in> ?S) ij" by auto
        qed
        with gGs have "\<exists> g \<in> set Gs. ?check g" by auto
      }
      hence main: "?R \<subseteq> (\<Union>G \<in> set Gs. sct_semantics.steps S (NS \<union> NST) ?conn G)" by auto
      have "SN ?R"
        by (rule sct_semantics.SCT_correctness2[where S = S and NS = "NS \<union> NST"], rule sct_semantics, rule main, rule check)
      obtain f where "f = ?pairs" by auto
      hence id: "?R = {(f k, f (Suc k)) | k. True}" by auto
      have "\<not> SN ?R" unfolding SN_defs by (simp only: id, blast) 
      with `SN ?R` show False by auto
    qed
  qed (auto simp: R P)
qed


lemma sct_with_subterm:
  assumes chain: "min_ichain (nfs,m,P, {}, Q, R, {}) s t \<sigma>"
    and tuple: "\<forall>(s, t)\<in>P. is_Fun s \<and> is_Fun t \<and> (the (root t)) \<notin> defs R"
    and graphs: "\<forall>(s, t)\<in>P. \<exists>stri nstri. (
      Scg (s, t, info s t) (s, t, info s t) stri nstri \<in> set Gs
      \<and> (\<forall>(i, j)\<in>set stri.
        i \<le> length (args s) \<and> j \<le> length (args t) \<and> (get_arg s i, get_arg t j) \<in> supt)
      \<and> (\<forall>(i, j)\<in>set nstri.
        i \<le> length (args s) \<and> j \<le> length (args t) \<and> (get_arg s i, get_arg t j) \<in> supteq))"
    and no_left_vars: "\<forall>(l, r)\<in>R. is_Fun l"
    and m: m
    and edg: "\<And> s t u v. ((s,t),(u,v)) \<in> DG nfs m P Q R \<Longrightarrow> edg (s,t,info s t) (u,v,info u v)"  
    and check: "check_SCT (\<lambda> (s,t,i_st) (u,v,i_uv). edg (s,t,i_st) (u,v,i_uv)) Gs" (is "check_SCT ?conn Gs")
   shows False
proof -
  let ?entry = "\<lambda>k. (s (Suc k), t (Suc k), info (s (Suc k)) (t (Suc k)))"
  let ?s = "\<lambda>i. s (Suc i)"
  let ?t = "\<lambda>i. t (Suc i)"
  let ?\<sigma> = "\<lambda>i. \<sigma> (Suc i)"
  let ?qrstep = "qrstep nfs Q R"
  let ?supt = "restrict_SN ((op \<rhd> \<union> ?qrstep)^+) ?qrstep"
  let ?supteq = "restrict_SN (?qrstep^* ) ?qrstep"
  let ?pairs = "\<lambda>k. (?entry k, (\<lambda>j. get_arg (?s k) j \<cdot> ?\<sigma> k))"
  let ?Rs = "?qrstep^*"
  let ?R = "{(?pairs k, ?pairs(Suc k)) | k. True}"
  let ?S = "(?supteq \<union> ?supt)^* O ?supt O (?supteq \<union> ?supt)^*"
  let ?NS = "(?supteq \<union> ?supt)^*"
  have sct_semantics: "sct_semantics ?supt ?supteq"
  proof
    show "SN ?supt"
    proof -
      have "\<forall>t. SN_on ?supt {t}"
      proof (rule ccontr)
        assume "\<not>(\<forall>t. SN_on ?supt {t})"
        then obtain t where "\<not> SN_on ?supt {t}" by auto
        then obtain S where "S 0 = t" and "\<forall>i. (S i, S (Suc i)) \<in> ?supt"
          unfolding SN_defs by auto
        hence "\<forall>i. SN_on ?qrstep {S i}" unfolding restrict_SN_def by simp
        hence "SN_on ?qrstep {S 0}" by simp
        hence "SN_on (op \<rhd> \<union> ?qrstep) {t}" unfolding `S 0 = t`
          by (rule SN_on_qrstep_imp_SN_on_supt_union_qrstep)
        hence "SN_on ((op \<rhd> \<union> ?qrstep)^+) {t}" by (rule SN_on_trancl)
        moreover have "?supt \<subseteq> (op \<rhd> \<union> ?qrstep)^+" by (rule restrict_SN_subset)
        ultimately have "SN_on ?supt {t}" by (rule SN_on_subset1)
        with `\<not> SN_on ?supt {t}` show False by simp
      qed
      thus ?thesis unfolding SN_defs by simp
    qed
  next
    show "?supteq O ?supt \<subseteq> ?supt"
    proof (rule subrelI)
      fix s t assume "(s, t) \<in> ?supteq O ?supt"
      then obtain u where fst: "(s, u) \<in> ?supteq" and snd: "(u, t) \<in> ?supt" by auto
      have "(s, u) \<in> ?Rs" and "SN_on ?qrstep {s}" using fst unfolding restrict_SN_def by auto
      moreover have "(u, t) \<in> (op \<rhd> \<union> ?qrstep)^+" and "SN_on ?qrstep {u}" using snd
        unfolding restrict_SN_def by auto
      ultimately show "(s, t) \<in> ?supt"
      proof (induct)
        case base thus ?case unfolding restrict_SN_def by simp
      next
        case (step v w)
        from `(v, w) \<in> ?qrstep` have "(v, w) \<in> ?qrstep^+" by simp
        with trancl_union_right[where r="?qrstep" and s="op \<rhd>"]
          have "(v, w) \<in> (op \<rhd> \<union> ?qrstep)^+" by blast
        with `(w, t) \<in> (op \<rhd> \<union> ?qrstep)^+` have "(v, t) \<in> (op \<rhd> \<union> ?qrstep)^+" by simp
        moreover have "SN_on ?qrstep {v}"
          by (rule steps_preserve_SN_on[OF `(s, v) \<in> ?Rs` `SN_on ?qrstep {s}`])
        ultimately show ?case using step by simp
      qed
    qed
  qed
  from chain `m` have min_t: "\<forall>i. SN_on ?qrstep {t i \<cdot> \<sigma> i}" by (auto simp: minimal_cond_def)
  from chain
    have inR: "\<forall>k. (t k \<cdot> \<sigma> k, s (Suc k) \<cdot> \<sigma> (Suc k)) \<in> ?qrstep^*" by (auto simp: ichain.simps)
  have "\<forall>i>(0::nat). SN_on ?qrstep {s i \<cdot> \<sigma> i}"
  proof (intro impI allI)
    fix i::nat assume "i > 0"
    then obtain n where "i = Suc n" by (cases i) auto
    from inR have step: "(t (i - 1) \<cdot> \<sigma> (i - 1), s i \<cdot> \<sigma> i) \<in> ?qrstep^*"
      unfolding `i = Suc n` by auto
    from min_t have "SN_on ?qrstep {t (i - 1) \<cdot> \<sigma> (i - 1)}" unfolding `i = Suc n` by auto
    from steps_preserve_SN_on[OF step this] show "SN_on ?qrstep {s i \<cdot> \<sigma> i}" .
  qed
  hence min_s: "\<forall>i. SN_on ?qrstep {?s i \<cdot> ?\<sigma> i}" by auto
  {
    fix k
    let ?sk = "\<lambda>i. get_arg (?s k) i \<cdot> ?\<sigma> k"
    let ?tk = "\<lambda>i. get_arg (?t k) i \<cdot> ?\<sigma> k"
    let ?sk1 = "\<lambda>i. get_arg (?s (Suc k)) i \<cdot> ?\<sigma> (Suc k)"
    from chain have inP: "(?s k, ?t k) \<in> P" by (auto simp: ichain.simps)
    from chain have inP': "(?s (Suc k), ?t (Suc k)) \<in> P" by (auto simp: ichain.simps)
    from chain have inR: "(?t k \<cdot> ?\<sigma> k, ?s (Suc k) \<cdot> ?\<sigma> (Suc k)) \<in> ?qrstep^*"
      by (auto simp: ichain.simps)
    have tk_supt: "\<forall>i\<le>length (args (?t k)). i > 0 \<longrightarrow> ?t k \<cdot> ?\<sigma> k \<rhd> ?tk i"
    proof (intro impI allI)
      fix i assume i: "i \<le> length (args (?t k))"
        and i2: "i > 0"
      thus "?t k \<cdot> ?\<sigma> k \<rhd> ?tk i"
      proof (cases "?t k")
        case (Var x) with tuple and inP show ?thesis by auto
      next
        case (Fun f ss)
        from i2 obtain j where j2: "i = Suc j" by (cases i, auto)
        with i have j: "j < length (args (?t k))" by auto 
        from j have "?t k \<rhd> args (?t k) ! j" unfolding Fun
          using supt.arg[of "args (Fun f ss) ! j" "ss" f] by auto
        thus ?thesis using j2 supt_stable[to_set] by auto
      qed
    qed
    have sk_supt: "\<forall>i\<le>length (args (?s k)). i > 0 \<longrightarrow> ?s k \<cdot> ?\<sigma> k \<rhd> ?sk i"
    proof (intro impI allI)
      fix i assume i: "i \<le> length (args (?s k))"
        and i2: "i > 0" thus "?s k \<cdot> ?\<sigma> k \<rhd> ?sk i"
      proof (cases "?s k")
        case (Var x) with tuple and inP show ?thesis by auto
      next
        case (Fun f ss)
        from i2 obtain j where j2: "i = Suc j" by (cases i, auto)
        with i have j: "j < length (args (?s k))" by auto 
        from j have "?s k \<rhd> args (?s k) ! j" unfolding Fun
          using supt.arg[of "args (Fun f ss) ! j" "ss" f] by auto
        thus ?thesis using j2 supt_stable[to_set] by auto
      qed
    qed
    from chain `m` have sntk: "SN_on ?qrstep {?t k \<cdot> ?\<sigma> k}" by (simp add: minimal_cond_def)
    have SN_tki: "\<forall>i \<le> length (args (?t k)). SN_on ?qrstep {?tk i}"
    proof
      fix i
      show "i \<le> length (args(?t k)) \<longrightarrow> SN_on ?qrstep {?tk i}"
        using subterm_preserves_SN_gen[OF ctxt_closed_qrstep sntk] tk_supt sntk
        by (cases i) auto
    qed
    have SN_sk: "SN_on ?qrstep {?s k \<cdot> ?\<sigma> k}" using min_s by simp
    have SN_ski: "\<forall>i\<le>length (args (?s k)). SN_on ?qrstep {?sk i}"
    proof
      fix i
      show "i \<le> length(args(?s k)) \<longrightarrow> SN_on ?qrstep {?sk i}"
        using subterm_preserves_SN_gen[OF ctxt_closed_qrstep SN_sk] sk_supt SN_sk
        by (cases i) auto
    qed
    note chain = chain[unfolded min_ichain.simps ichain.simps minimal_cond_def]
    from inP tuple
      have validDP: "is_Fun (?t k) \<and> (the (root (?t k))) \<notin> defs R"
      by auto
    from chain inP tuple have validDP1: "is_Fun (?s (Suc k))" by auto
    from chain have steps: "(?t k \<cdot> ?\<sigma> k, ?s(Suc k) \<cdot> ?\<sigma>(Suc k)) \<in> ?Rs" by auto
    have stepsNS: "\<forall>j\<le>length (args (?t k)). (?tk j, ?sk1 j) \<in> ?NS"
    proof (intro allI impI)
      fix j1
      assume j1: "j1 \<le> length (args (?t k))"
      show "(?tk j1, ?sk1 j1) \<in> ?NS"
      proof (cases j1)
        case 0
        with steps show ?thesis unfolding restrict_SN_def using sntk by auto
      next
        case (Suc j)
        let ?tj = "args (?t k) ! j"
        let ?sj = "args (?s (Suc k)) ! j"
        from j1 Suc have j: "j < length (args (?t k))" by simp
        from validDP obtain f ts
          where tk: "?t k = Fun f ts \<and> (f, length ts) \<notin> defs R" by (cases "?t k") auto
        from validDP1 obtain g ss
          where sk1: "?s (Suc k) = Fun g ss" by (cases "?s (Suc k)") auto
        let ?tss = "map (\<lambda>t. t \<cdot> ?\<sigma> k) ts"
        let ?sss = "map (\<lambda>s. s \<cdot> ?\<sigma> (Suc k)) ss"
        from tk have f: "\<not> defined R (f, length ?tss)" by (simp add: defs_def)
        from steps have "(Fun f ts \<cdot> ?\<sigma> k, Fun g ss \<cdot> ?\<sigma> (Suc k)) \<in> ?Rs" using tk sk1 by auto
        hence steps2: "(Fun f ?tss, Fun g ?sss) \<in> ?Rs" by auto
        from this no_left_vars f
          have "\<exists>us. length us = length ?tss
            \<and> Fun g ?sss = Fun f us
            \<and> (\<forall>i<length ?tss. (?tss ! i, us ! i) \<in> ?Rs)"
          by (rule nondef_root_imp_arg_qrsteps)
        from this obtain us
          where nearly: "length us = length ?tss
            \<and> Fun g ?sss = Fun f us
            \<and> (\<forall>i<length ?tss. (?tss ! i, us ! i) \<in> ?Rs)" ..
        from SN_tki have SN_tkj: "SN_on ?qrstep {?tj \<cdot> ?\<sigma> k}" using j1 Suc by auto
        from nearly j tk sk1 have "(?tj \<cdot> ?\<sigma> k, ?sj \<cdot> ?\<sigma> (Suc k)) \<in> ?Rs" by auto
        with SN_tkj have "(?tj \<cdot> ?\<sigma> k, ?sj \<cdot> ?\<sigma> (Suc k)) \<in> restrict_SN ?Rs ?qrstep"
          unfolding restrict_SN_def by simp
        thus ?thesis using Suc unfolding restrict_SN_def by auto
      qed
    qed
    have DG: "((?s k, ?t k), (?s (Suc k), ?t (Suc k))) \<in> DG nfs m P Q R"
      by (rule DG_I[OF inP inP' steps _ _ _ _ sntk], insert chain, auto)

    from edg[OF this]
      have conn: "?conn (?entry k) (?entry (Suc k))" by auto
    from inP obtain stri nstri
      where graph: "Scg (?entry k) (?entry k) stri nstri \<in> set Gs
        \<and> (\<forall>(i, j)\<in>set stri.
          i \<le> length (args (?s k)) \<and> j \<le> length (args (?t k))
          \<and> get_arg (?s k) i \<rhd> get_arg (?t k) j)
        \<and> (\<forall>(i, j)\<in>set nstri.
          i \<le> length (args (?s k)) \<and> j \<le> length (args (?t k))
          \<and> get_arg (?s k) i \<unrhd> get_arg (?t k) j)" using graphs by force
    let ?g = "Scg (?entry k) (?entry k) stri nstri"
    from graph have gGs: "?g \<in> set Gs" ..
    let ?check = "\<lambda>g. ((?entry k, ?sk), (?entry (Suc k), ?sk1))
           \<in> sct_semantics.steps ?supt ?supteq ?conn g"
    have "?check ?g"
    proof (simp add: sct_semantics.steps.simps sct_semantics conn[simplified], rule conjI)
      {
        fix i j
        assume "(i, j) \<in> set nstri"
        with graph have i: "i \<le> length(args(?s k))" and j: "j \<le> length(args(?t k))"
          and NS: "get_arg(?s k) i \<unrhd> get_arg (?t k) j" by auto
        from j stepsNS have astepsNS: "(?tk j, ?sk1 j) \<in> ?NS" by auto
        from SN_ski and i have snski: "SN_on ?qrstep {?sk i}" by simp
        from NS supteq_subst_closed have "?sk i \<unrhd> ?tk j"
          unfolding subst_closed_def using subst_closureI by blast
        hence non_strict: "(?sk i, ?tk j) \<in> ?NS" using snski
          unfolding restrict_SN_def supteq_supt_conv by force
        from rtrancl_trans[OF non_strict astepsNS] have "(?sk i, ?sk1 j) \<in> ?NS" by auto
      }
      thus "\<forall>ij\<in>set nstri.(\<lambda>(i, j). (?sk i, ?sk1 j) \<in> ?NS) ij" by auto
    next
      {
        fix i j
        assume "(i, j) \<in> set stri"
        with graph have i: "i \<le> length (args(?s k))" and j: "j \<le> length(args(?t k))"
          and S: "get_arg(?s k) i \<rhd> get_arg(?t k) j" by auto
        from j stepsNS have astepsNS: "(?tk j, ?sk1 j) \<in> ?NS" by auto
        from SN_ski and i have snski: "SN_on ?qrstep {?sk i}" by simp
        from S supt_subst_closed have "?sk i \<rhd> ?tk j"
          unfolding subst_closed_def using subst_closureI by blast
        hence strict: "(?sk i, ?tk j) \<in> ?supt" using snski
          unfolding restrict_SN_def by force
        have "(?sk i, ?sk1 j) \<in> ?S" using strict astepsNS by auto
      }
      thus "\<forall>ij\<in>set stri.(\<lambda>(i, j). (?sk i, ?sk1 j) \<in> ?S) ij" by auto
    qed
    with gGs have "\<exists>g\<in>set Gs. ?check g" by blast
  }
  hence main: "?R \<subseteq> (\<Union>G\<in>set Gs. sct_semantics.steps ?supt ?supteq ?conn G)" by auto
  have "SN ?R"
    by (rule sct_semantics.SCT_correctness2[where S = ?supt and NS = ?supteq],rule sct_semantics,rule main,rule check)
  obtain f where "f = ?pairs" by auto
  hence id: "?R = {(f k, f (Suc k)) | k. True}" by auto
  have "\<not> SN ?R" unfolding SN_defs by (simp only: id, blast)
  with `SN ?R` show False by auto
qed

end

